import numpy as np
import os
import pytest
import tempfile

from ultranest.mlfriends import ScalingLayer, AffineLayer, MLFriends
from ultranest import ReactiveNestedSampler
from ultranest.stepsampler import RegionMHSampler, CubeMHSampler, SliceSampler, CubeSliceSampler, RegionSliceSampler, SpeedVariableRegionSliceSampler, RegionBallSliceSampler, SpeedVariableGenerator
from ultranest.stepsampler import ellipsoid_bracket, crop_bracket_at_unit_cube, _inside_region
from ultranest.stepsampler import generate_random_direction, generate_cube_oriented_direction
from ultranest.stepsampler import SequentialDirectionGenerator, OrthogonalDirectionGenerator, SequentialRegionDirectionGenerator
from ultranest.stepsampler import generate_region_random_direction, generate_region_oriented_direction, generate_cube_oriented_differential_direction
from ultranest.stepsampler import generate_differential_direction, generate_partial_differential_direction, generate_mixture_random_direction

from ultranest.pathsampler import SamplingPathStepSampler
from ultranest.stepsampler import select_random_livepoint, IslandPopulationRandomLivepointSelector
from numpy.testing import assert_allclose

#here = os.path.dirname(__file__)

def loglike_vectorized(z):
    a = np.array([-0.5 * sum([((xi - 0.7 + i*0.001)/0.1)**2 for i, xi in enumerate(x)]) for x in z])
    b = np.array([-0.5 * sum([((xi - 0.3 - i*0.001)/0.1)**2 for i, xi in enumerate(x)]) for x in z])
    return np.logaddexp(a, b)

def loglike(x):
    a = -0.5 * sum([((xi - 0.7 + i*0.001)/0.1)**2 for i, xi in enumerate(x)])
    b = -0.5 * sum([((xi - 0.3 - i*0.001)/0.1)**2 for i, xi in enumerate(x)])
    return np.logaddexp(a, b)

def transform(x):
    return x # * 10. - 5.

def transform1(x):
    return x**2

paramnames = ['param%d' % i for i in range(3)]
#paramnames = ['param%d' % i for i in range(40)]

def test_stepsampler_cubemh(plot=False):
    np.random.seed(1)
    sampler = ReactiveNestedSampler(paramnames, loglike_vectorized, transform=transform1, vectorized=True)
    sampler.stepsampler = CubeMHSampler(nsteps=4 * len(paramnames))
    r = sampler.run(log_interval=50, min_num_live_points=400)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1, a.sum()
    assert b.sum() > 1, b.sum()

    # check that diagnostics fail
    print("mean jump distance:", sampler.stepsampler.mean_jump_distance)
    print("far enough fraction:", sampler.stepsampler.far_enough_fraction)
    assert sampler.stepsampler.mean_jump_distance < 1.0, sampler.stepsampler.mean_jump_distance
    assert sampler.stepsampler.far_enough_fraction < 0.5, sampler.stepsampler.far_enough_fraction

    print("Diagnostic print:")
    sampler.stepsampler.print_diagnostic()

def test_stepsampler_regionmh(plot=False):
    np.random.seed(2)
    sampler = ReactiveNestedSampler(paramnames, loglike_vectorized, transform=transform, vectorized=True)
    sampler.stepsampler = RegionMHSampler(nsteps=4 * len(paramnames))
    r = sampler.run(log_interval=50, min_num_live_points=400)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1, a
    assert b.sum() > 1, b

def test_direction_proposals():
    ndim = 10
    np.random.seed(12)
    region = make_region(ndim)
    ui = region.u[0]

    scale = np.random.uniform()
    vcube = generate_cube_oriented_direction(ui, region, scale)
    assert (vcube != 0).sum() == 1, vcube
    assert np.linalg.norm(vcube) == scale, vcube

    vcubede = generate_cube_oriented_differential_direction(ui, region, scale)
    assert (vcubede != 0).sum() == 1, vcubede
    assert np.linalg.norm(vcubede) > 0, vcubede

    vharm = generate_random_direction(ui, region, scale)
    assert (vharm != 0).all(), vharm

    vde = generate_differential_direction(ui, region, scale)
    assert (vde != 0).all(), vde

    vregionslice = generate_region_oriented_direction(ui, region, scale)
    assert (vregionslice != 0).all(), vregionslice

    vmix = generate_mixture_random_direction(ui, region, scale)
    assert (vmix != 0).all(), vmix

    vregionharm = generate_region_random_direction(ui, region, scale)
    assert (vregionharm != 0).all(), vregionharm

    direction_generator = SequentialDirectionGenerator()
    for i in range(ndim * 2):
        vdir = direction_generator(ui, region, scale)
        assert (vdir != 0).sum() == 1, vdir
        assert np.abs(vdir[i % ndim]) > 0, vdir

    region_direction_generator = SequentialRegionDirectionGenerator()
    for i in range(ndim * 2):
        vdirharm = region_direction_generator(ui, region, scale)
        assert (vdirharm != 0).all(), vdirharm

    vpartialde = generate_partial_differential_direction(ui, region, scale)
    assert (vpartialde != 0).sum() > 1, vpartialde
    assert (vpartialde != 0).sum() < ndim, vpartialde

    # test that applying OrthogonalDirectionGenerator to SequentialDirectionGenerator has no effect
    ortho_direction_generator = OrthogonalDirectionGenerator(SequentialDirectionGenerator())
    for i in range(ndim * 2):
        vdir = ortho_direction_generator(ui, region, scale)
        assert (vdir != 0).sum() == 1, vdir
        assert np.abs(vdir[i % ndim]) > 0, vdir

def test_inside_region():
    ndim = 10
    np.random.seed(12)
    region = make_region(ndim, us = np.random.uniform(0.5, 0.51, size=(400, ndim)))
    i = np.random.randint(400)
    ui = region.u[i]
    assert _inside_region(region, ui, ui)
    # corner case where a new point is close to a old case, but both are somehow outside the region
    unew, uold = np.random.uniform(0.4, 0.401, size=(2, ndim))
    assert _inside_region(region, unew, uold)

def test_stepsampler_cubeslice(plot=False):
    np.random.seed(3)
    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform1)
    sampler.stepsampler = CubeSliceSampler(nsteps=len(paramnames))
    r = sampler.run(log_interval=50, min_num_live_points=400)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1
    assert b.sum() > 1

def test_stepsampler_regionslice(plot=False):
    np.random.seed(4)
    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform)
    sampler.stepsampler = RegionSliceSampler(nsteps=2 + len(paramnames))
    r = sampler.run(log_interval=50, min_num_live_points=400)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1
    assert b.sum() > 1

    # check that diagnostics pass
    print("mean jump distance:", sampler.stepsampler.mean_jump_distance)
    print("far enough fraction:", sampler.stepsampler.far_enough_fraction)
    assert sampler.stepsampler.mean_jump_distance > 1.0, sampler.stepsampler.mean_jump_distance
    assert sampler.stepsampler.far_enough_fraction > 0.6, sampler.stepsampler.far_enough_fraction

    print("Diagnostic print:")
    sampler.stepsampler.print_diagnostic()

def test_SpeedVariableGenerator():
    np.random.seed(4)
    ndims = [3, 10]
    matrices = [
        np.array([[True, True, True], [False, True, True], [False, False, True]]),
        [Ellipsis, slice(1,None), slice(2,4)]
    ]
    for matrix, ndim in zip(matrices, ndims):
        direction_generator = SpeedVariableGenerator(matrix, generate_direction=generate_random_direction)
        for i in range(10):
            u0 = np.random.uniform(size=ndim)
            for mask_varying in matrix:
                mask = np.zeros(ndim, dtype=bool)
                mask[mask_varying] = True
                print("starting at u0", u0)
                print("varying:", mask_varying, mask)
                v = direction_generator(u0, None)
                print("direction:", v)
                assert_allclose(v[~mask], 0)
                u1 = u0 + np.random.uniform() * v
                print("new point:", u1)
                assert_allclose(u1[~mask], u0[~mask])


def test_stepsampler_variable_speed_SLOW(plot=False):
    matrices = [
        np.array([[True, True, True], [False, True, True], [False, False, True]]),
        [Ellipsis, slice(1,None), slice(2,4)]
    ]
    for i, matrix in enumerate(matrices):
        np.random.seed(42 + i)
        sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform)
        sampler.stepsampler = SpeedVariableRegionSliceSampler(matrix)
        r = sampler.run(log_interval=50, min_num_live_points=400)
        sampler.print_results()
        a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
        b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
        assert a.sum() > 1
        assert b.sum() > 1

def make_region(ndim, us=None):
    if us is None:
        us = np.random.uniform(size=(1000, ndim))

    if ndim > 1:
        transformLayer = AffineLayer()
    else:
        transformLayer = ScalingLayer()
    transformLayer.optimize(us, us)
    region = MLFriends(us, transformLayer)
    region.maxradiussq, region.enlarge = region.compute_enlargement(nbootstraps=30)
    region.create_ellipsoid(minvol=1.0)
    return region


def test_stepsampler(plot=False):
    np.random.seed(6)
    region = make_region(len(paramnames))
    Ls = loglike_vectorized(region.u)

    stepsampler = CubeMHSampler(nsteps=len(paramnames))
    while True:
        u1, p1, L1, nc = stepsampler.__next__(region, -1e100, region.u, Ls, transform, loglike)
        if u1 is not None:
            break
    assert L1 > -1e100
    print(u1, L1)
    while True:
        u2, p2, L2, nc = stepsampler.__next__(region, -1e100, region.u, Ls, transform, loglike)
        if u2 is not None:
            break
    assert L2 > -1e100
    print(u2, L2)
    assert np.all(u1 != u2)
    assert np.all(L1 != L2)

def test_stepsampler_adapt_when_stuck(plot=False):
    # check that a stuck sampler can free itself
    np.random.seed(7)
    us = np.random.normal(0.7, 0.0001, size=(1000, len(paramnames)))
    region = make_region(len(paramnames), us=us)
    Ls = loglike_vectorized(us)
    Lmin = Ls.min()

    print('CubeMHSampler')
    stepsampler = CubeMHSampler(nsteps=1, region_filter=True)
    np.random.seed(23)
    old_scale = stepsampler.scale
    for i in range(1000):
        if i > 100:
            assert False, i
        unew, pnew, Lnew, nc = stepsampler.__next__(region, Lmin, us, Ls, transform, loglike, ndraw=10)
        if unew is not None:
            break

    new_scale = stepsampler.scale
    assert new_scale != old_scale
    assert new_scale < 0.01, (new_scale, unew)

    print('CubeSliceSampler')
    stepsampler = SliceSampler(nsteps=1, region_filter=True, generate_direction=generate_cube_oriented_direction)
    np.random.seed(23)
    old_scale = stepsampler.scale
    for j in range(100):
        for i in range(1000):
            if i > 100:
                assert False, i
            unew, pnew, Lnew, nc = stepsampler.__next__(region, Lmin, us, Ls, transform, loglike, ndraw=10)
            if unew is not None:
                break

    new_scale = stepsampler.scale
    assert new_scale != old_scale
    assert new_scale < 0.01, (new_scale, unew)

def test_stepsampler_adapt(plot=True):
    np.random.seed(8)
    region = make_region(len(paramnames))
    Ls = loglike_vectorized(region.u)
    with pytest.raises(ValueError):
        RegionMHSampler(nsteps=len(paramnames), adaptive_nsteps='Hello')

    with tempfile.TemporaryDirectory() as tempdir:
        for sampler_class in RegionMHSampler, CubeMHSampler, CubeSliceSampler, RegionSliceSampler:
            for adaptation in False, 'move-distance', 'move-distance-midway', 'proposal-total-distances', 'proposal-summed-distances':
                print()
                if sampler_class in (CubeMHSampler, CubeSliceSampler):
                    logfilename = os.path.join(tempdir, 'test-stepsampler-%s.log' % adaptation)
                    log = open(logfilename, 'w')
                else:
                    logfilename = None
                    log = False
                stepsampler = sampler_class(nsteps=len(paramnames), adaptive_nsteps=adaptation, log=log)
                print(stepsampler)
                stepsampler.region_changed(Ls, region)
                np.random.seed(23)
                old_scale = stepsampler.scale
                for i in range(5):
                    while True:
                        unew, pnew, Lnew, nc = stepsampler.__next__(region, -1e100, region.u, Ls, transform, loglike)
                        if unew is not None:
                            break
                new_scale = stepsampler.scale
                assert new_scale != old_scale

                if adaptation:
                    assert stepsampler.nsteps != len(paramnames)
                else:
                    assert stepsampler.nsteps == len(paramnames)

                if logfilename:
                    print(np.loadtxt(logfilename).shape)
                    log_nentries, log_ncolumns = np.loadtxt(logfilename).shape
                    assert log_nentries == 5
                    assert log_ncolumns == (1 + 4 * len(unew) + 7)

                if adaptation == 'move-distance' and sampler_class == RegionSliceSampler and plot:
                    # test plotting
                    prefix = os.path.join(tempdir, 'test-stepsampler')
                    assert not os.path.exists(prefix + '-plot.pdf')
                    stepsampler.plot(prefix + '-plot.pdf')
                    assert os.path.exists(prefix + '-plot.pdf')

                    assert not os.path.exists(prefix + '-plot-jumps.pdf')
                    stepsampler.plot_jump_diagnostic_histogram(prefix + '-plot-jumps.pdf')
                    assert os.path.exists(prefix + '-plot-jumps.pdf')

def assert_point_touches_ellipsoid(ucurrent, v, t, ellipsoid_center, ellipsoid_invcov, enlarge):
    unext = ucurrent + v * t
    d = unext - ellipsoid_center
    r = np.einsum('j,jk,k->', d, ellipsoid_invcov, d)
    assert np.isclose(r, enlarge), (ucurrent, t, r, enlarge)

def test_ellipsoid_bracket(plot=False):
    for seed in range(20):
        print("seed:", seed)
        np.random.seed(seed)
        if seed % 2 == 0:
            us = np.random.normal(size=(2**np.random.randint(3, 10), 2))
            us /= ((us**2).sum(axis=1)**0.5).reshape((-1, 1))
            us = us * 0.1 + 0.5
        else:
            us = np.random.uniform(size=(2**np.random.randint(3, 10), 2))

        if plot:
            import matplotlib.pyplot as plt
            plt.plot(us[:,0], us[:,1], 'o ', ms=2)

        transformLayer = ScalingLayer()
        region = MLFriends(us, transformLayer)
        try:
            region.maxradiussq, region.enlarge = region.compute_enlargement()
            region.create_ellipsoid()
        except ValueError:
            continue

        print(region.ellipsoid_center)
        print(region.enlarge)
        print(region.ellipsoid_cov)
        print(region.ellipsoid_invcov)
        print(region.ellipsoid_axes)
        print(region.ellipsoid_inv_axes)

        ucurrent = np.array([2**0.5*0.1/2+0.5, 2**0.5*0.1/2+0.5])
        ucurrent = np.array([0.4, 0.525])
        v = np.array([1., 0])
        if plot: plt.plot(ucurrent[0], ucurrent[1], 'o')
        print("from", ucurrent, "in direction", v)
        left, right = ellipsoid_bracket(ucurrent, v, region.ellipsoid_center, region.ellipsoid_inv_axes, region.enlarge)
        uleft = ucurrent + v * left
        uright = ucurrent + v * right

        if plot:
            plt.plot([uleft[0], uright[0]], [uleft[1], uright[1]], 'x-')

            plt.savefig('test_ellipsoid_bracket.pdf', bbox_inches='tight')
            plt.close()
        print("ellipsoid bracket:", left, right)
        assert left <= 0, left
        assert right >= 0, right

        assert_point_touches_ellipsoid(ucurrent, v, left, region.ellipsoid_center, region.ellipsoid_invcov, region.enlarge)
        assert_point_touches_ellipsoid(ucurrent, v, right, region.ellipsoid_center, region.ellipsoid_invcov, region.enlarge)

def test_crop_bracket(plot=False):
    ucurrent = np.array([0.39676747, 0.53881673])
    v = np.array([-0.79619985, -0.60503372])
    ellipsoid_center = np.array([0.23556461, 0.49899689])
    ellipsoid_inv_axes = np.array([[-3.28755896,  0.70136518], [ 1.33333377,  1.72933397]])
    enlarge = 26.66439694551674
    ellipsoid_invcov = np.array([[11.29995701, -3.17051875], [-3.17051875,  4.76837493]])
    #enlarge = 1.0
    #ellipsoid_inv_axes = np.array([[1.0, 0.], [0., 1]])


    eleft, eright = ellipsoid_bracket(ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    if plot:
        us = np.random.uniform(-2, +2, size=(10000, 2))
        d = us - ellipsoid_center
        r = np.einsum('ij,jk,ik->i', d, ellipsoid_invcov, d)
        mask_inside = r <= enlarge

        import matplotlib.pyplot as plt
        plt.plot(us[mask_inside,0], us[mask_inside,1], '+', ms=2)
        plt.plot(ucurrent[0], ucurrent[1], 'o ', ms=2)
        plt.plot([ucurrent[0] + eleft * v[0], ucurrent[0] + eright * v[0]],
            [ucurrent[1] + eleft * v[1], ucurrent[1] + eright * v[1]],
            '-s', ms=8)
        plt.savefig('test_crop_bracket.pdf', bbox_inches='tight')
    print("left:", eleft, ucurrent + v * eleft)
    assert eleft <= 0, eleft
    assert_point_touches_ellipsoid(ucurrent, v, eleft, ellipsoid_center, ellipsoid_invcov, enlarge)
    print("right:", eright, ucurrent + v * eright)
    assert eright >= 0, eright
    assert_point_touches_ellipsoid(ucurrent, v, eright, ellipsoid_center, ellipsoid_invcov, enlarge)

    left, right, cropleft, cropright = crop_bracket_at_unit_cube(ucurrent, v, eleft, eright)
    if plot:
        plt.plot([ucurrent[0] + left * v[0], ucurrent[0] + right * v[0]],
            [ucurrent[1] + left * v[1], ucurrent[1] + right * v[1]],
            's--', ms=8)
        plt.savefig('test_crop_bracket.pdf', bbox_inches='tight')
        plt.close()
    assert cropleft
    assert cropright
    assert (ucurrent + v * left <= 1).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    assert (ucurrent + v * right <= 1).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    assert (ucurrent + v * left >= 0).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    assert (ucurrent + v * right >= 0).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)

    left, right, cropleft, cropright = crop_bracket_at_unit_cube(ucurrent, -v, eleft, eright)
    if plot:
        plt.plot([ucurrent[0] - left * v[0], ucurrent[0] - right * v[0]],
            [ucurrent[1] - left * v[1], ucurrent[1] - right * v[1]],
            's--', ms=8)
        plt.savefig('test_crop_bracket_negative.pdf', bbox_inches='tight')
        plt.close()
    assert cropleft
    assert cropright
    assert (ucurrent - v * left <= 1).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    assert (ucurrent - v * right <= 1).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    assert (ucurrent - v * left >= 0).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)
    assert (ucurrent - v * right >= 0).all(), (ucurrent, v, ellipsoid_center, ellipsoid_inv_axes, enlarge)

def test_random_point_selector():
    np.random.seed(41)
    K = 10
    ndim = 2
    i1 = np.random.randint(0, K)
    i2 = np.random.randint(0, K)
    i3 = np.random.randint(0, K)
    us = np.random.normal(size=(K, ndim))
    Ls = np.random.normal(size=K)
    Lmin = Ls.min()
    np.random.seed(41)
    j1 = select_random_livepoint(us, Ls, Lmin)
    j2 = select_random_livepoint(us, Ls, Lmin)
    j3 = select_random_livepoint(us, Ls, Lmin)
    assert i1 == j1, (i1, j1)
    assert i2 == j2, (i2, j2)
    assert i3 == j3, (i3, j3)


def test_island_point_selector():
    K = 10
    ndim = 2
    self_selector = IslandPopulationRandomLivepointSelector(1)
    selector = IslandPopulationRandomLivepointSelector(5)
    imbalanced_selector = IslandPopulationRandomLivepointSelector(9)
    for i in range(100):
        us = np.random.normal(size=(K, ndim))
        Ls = np.random.normal(size=K)
        Lmin = Ls.min()
        j1 = np.argmin(Ls)
        j2 = selector(us, Ls, Lmin)
        assert j1 == self_selector(us, Ls, Lmin)
        if j1 >= 5:
            assert j2 >= 5, (j1, j2)
        if j1 < 5:
            assert j2 < 5, (j1, j2)
        if j1 == 9:
            assert j1 == imbalanced_selector(us, Ls, Lmin)
        else:
            assert imbalanced_selector(us, Ls, Lmin) < 9

    np.random.seed(421)
    leaked = False
    selector = IslandPopulationRandomLivepointSelector(5, 0.1)
    for i in range(100):
        j1 = np.argmin(Ls)
        j2 = selector(us, Ls, Lmin)
        if j1 >= 5 and j2 >= 5 or j1 < 5 and j2 < 5:
            pass
        else:
            # leak, as expected
            leaked = True
            break
    assert leaked


if __name__ == '__main__':
    #test_stepsampler_cubemh(plot=True)
    #test_stepsampler_regionmh(plot=False)
    #test_stepsampler_de(plot=False)
    #test_stepsampler_cubeslice(plot=True)
    #test_stepsampler_regionslice(plot=True)
    test_ellipsoid_bracket()
    #test_crop_bracket(plot=True)



from sklearn.neural_network import MLPClassifier
import numpy as np
from mlfriends import RobustEllipsoidRegion


def filter_points(root, pointpile, ellipsoid_region):
    """Find all stored points, live or dead, that are within ellipsoid."""
    ids_inside = ellipsoid_region.inside_ellipsoid(pointpile.us)
    N = ids_inside.sum()
    found_i = 0
    found_us = pointpile.us[ids_inside,:]
    found_Ls = np.empty((N,))
    open_nodes = list(root.children)
    while len(open_nodes) > 0:
        n = open_nodes.pop(0)
        if ids_inside[n.id]:
            found_Ls[found_i] = n.value
            found_i += 1
        open_nodes += n.children
    assert found_i == N
    return found_us, found_Ls

# use viz_callback to build a neural network
# enable vectorization and set ndraw_min high.

class NeuralFilter(RobustEllipsoidRegion):
    def __init__(self, sampler, viz_callback, prob_cutoff, num_networks=4,
            neural_network_kwargs={}, verbose=True):

        default_neural_network_kwargs = dict(
            hidden_layer_sizes=(100, 50, 20), alpha=0, learning_rate_init=1e-2,
            max_iter=10000, tol=0, n_iter_no_change=10)
        default_neural_network_kwargs.update(neural_network_kwargs)
        self.neural_network_kwargs = default_neural_network_kwargs

        self.sampler = sampler
        self.viz_callback = viz_callback
        self.num_networks = num_networks
        self.prob_cutoff = prob_cutoff
        self.verbose = verbose
        self.networks = []

    def train(self, points, region, transformLayer):
        # get all stored/pile points within ellipsoid
        if self.verbose:
            print("training: getting data...")
        us, Ls = filter_points(self.sampler.root, self.sampler.pointpile, region)
        # transform to a whitened space with ellipsoid
        ts = transformLayer.transform(us)
        # then train neural network classifier to distinguish rejected from live points
        # classification training data:
        inside_class = (Ls > self.sampler.Lmin) * 1
        self.networks = []
        for i in range(self.num_networks):
            if self.verbose:
                print("training[%d/%d] with %d/%d samples ..." % (i+1, self.num_networks, inside_class.sum(), len(inside_class)))
            network = MLPClassifier(**self.neural_network_kwargs)
            network.fit(ts, inside_class)
            self.networks.append((transformLayer, network))

    def viz_callback(self, points, info, region, transformLayer, **kwargs):
        self.train(points=points, region=region, transformLayer=transformLayer)
        return self.viz_callback(points=points, info=info, transformLayer=transformLayer, **kwargs)

    def inside(self, us):
        proba = [network.predict_proba(layer.transform(us))[:,1]
            for layer, network in self.networks]
        mean_proba = np.mean(proba, axis=1)
        std_proba = np.std(proba, axis=1).mean()
        is_inside = mean_proba > self.prob_cutoff
        if self.verbose:
            print("prediction frac=%.3f%% mean[prob]=%.3f, std[cross-NN]=%.3f" % (is_inside.mean() * 100, mean_proba.mean(), std_proba))
        
        return is_inside

    def sample_from_boundingbox(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from bounding box around region.

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # draw from unit cube in prior space
        u = np.random.uniform(size=(nsamples, ndim))
        wmask = self.inside(u)
        return u[wmask,:]

    def sample_from_transformed_boundingbox(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from bounding box around region (in whitened space).

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # draw from rectangle in transformed space
        v = np.random.uniform(self.bbox_lo - self.maxradiussq, self.bbox_hi + self.maxradiussq, size=(nsamples, ndim))

        # check if inside unit cube
        w = self.transformLayer.untransform(v)
        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        wmask[wmask] = self.inside(w[wmask])

        return w[wmask,:]

    def sample_from_wrapping_ellipsoid(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from wrapping ellipsoid and filters with region.

        Parameters as described in ``sample()``.
        """
        N, ndim = self.u.shape
        # draw from rectangle in transformed space

        z = np.random.normal(size=(nsamples, ndim))
        assert ((z**2).sum(axis=1) > 0).all(), (z**2).sum(axis=1)
        z /= ((z**2).sum(axis=1)**0.5).reshape((nsamples, 1))
        assert self.enlarge > 0, self.enlarge
        u = z * self.enlarge**0.5 * np.random.uniform(size=(nsamples, 1))**(1./ndim)

        w = self.ellipsoid_center + np.dot(u, self.ellipsoid_axes_T)
        #assert self.inside_ellipsoid(w).all()

        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        wmask[wmask] = self.inside(w[wmask])
        return w[wmask,:]



import os
import tempfile
import numpy as np

from ultranest import ReactiveNestedSampler
from ultranest.mlfriends import AffineLayer, ScalingLayer, MLFriends, RobustEllipsoidRegion, SimpleRegion
from ultranest.popstepsampler import PopulationSliceSampler, PopulationRandomWalkSampler
from ultranest.popstepsampler import generate_cube_oriented_direction, generate_random_direction, generate_cube_oriented_direction_scaled
from ultranest.popstepsampler import generate_region_oriented_direction, generate_region_random_direction

def make_region(ndim, us=None, nlive=400):
    if us is None:
        us = np.random.uniform(size=(nlive, ndim))
    
    if ndim > 1:
        transformLayer = AffineLayer()
    else:
        transformLayer = ScalingLayer()
    transformLayer.optimize(us, us)
    region = MLFriends(us, transformLayer)
    region.maxradiussq, region.enlarge = region.compute_enlargement(nbootstraps=30)
    region.create_ellipsoid(minvol=1.0)
    return region

def loglike_vectorized(z):
    a = np.array([-0.5 * sum([((xi - 0.7 + i*0.001)/0.1)**2 for i, xi in enumerate(x)]) for x in z])
    b = np.array([-0.5 * sum([((xi - 0.3 - i*0.001)/0.1)**2 for i, xi in enumerate(x)]) for x in z])
    return np.logaddexp(a, b)

def loglike(x):
    a = -0.5 * sum([((xi - 0.7 + i*0.001)/0.1)**2 for i, xi in enumerate(x)])
    b = -0.5 * sum([((xi - 0.3 - i*0.001)/0.1)**2 for i, xi in enumerate(x)])
    return np.logaddexp(a, b)

def transform(x):
    return x # * 10. - 5.

paramnames = ['param%d' % i for i in range(3)]

def test_stepsampler_cubeslice(plot=False):
    np.random.seed(3)
    nsteps = np.random.randint(10, 50)
    popsize = np.random.randint(1, 20)
    sampler = ReactiveNestedSampler(paramnames, loglike_vectorized, transform=transform, vectorized=True)

    sampler.stepsampler = PopulationSliceSampler(
        popsize=popsize, nsteps=nsteps, 
        generate_direction=generate_cube_oriented_direction,
        log=True,
    )
    r = sampler.run(viz_callback=None, log_interval=50)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1
    assert b.sum() > 1

    with tempfile.TemporaryDirectory() as tempdir:
        prefix = os.path.join(tempdir, 'test-stepsampler')
        sampler.stepsampler.plot(prefix + '-plot.pdf')
        assert os.path.exists(prefix + '-plot.pdf')
        sampler.stepsampler.plot_jump_diagnostic_histogram(prefix + '-plot-jumps.pdf')
        assert os.path.exists(prefix + '-plot-jumps.pdf')
        sampler.stepsampler.print_diagnostic()
        print(sampler.stepsampler)
        print(sampler.stepsampler.status)

def test_stepsampler_cubegausswalk(plot=False):
    np.random.seed(2)
    nsteps = np.random.randint(10, 50)
    popsize = np.random.randint(1, 20)
    sampler = ReactiveNestedSampler(paramnames, loglike_vectorized, transform=transform, vectorized=True)

    sampler.stepsampler = PopulationRandomWalkSampler(
        popsize=popsize, nsteps=nsteps, 
        generate_direction=generate_cube_oriented_direction,
        scale=0.1, log=True,
    )
    r = sampler.run(viz_callback=None, log_interval=50, max_iters=200, max_num_improvement_loops=0)
    sampler.print_results()
    a = (np.abs(r['samples'] - 0.7) < 0.1).all(axis=1)
    b = (np.abs(r['samples'] - 0.3) < 0.1).all(axis=1)
    assert a.sum() > 1
    assert b.sum() > 1

    with tempfile.TemporaryDirectory() as tempdir:
        prefix = os.path.join(tempdir, 'test-stepsampler')
        sampler.stepsampler.plot(prefix + '-plot.pdf')
        assert os.path.exists(prefix + '-plot.pdf')
        sampler.stepsampler.plot_jump_diagnostic_histogram(prefix + '-plot-jumps.pdf')
        assert os.path.exists(prefix + '-plot-jumps.pdf')
        sampler.stepsampler.print_diagnostic()
        print(sampler.stepsampler)

def test_direction_proposals():
    proposals = [generate_cube_oriented_direction, generate_random_direction, 
        generate_region_oriented_direction, generate_region_random_direction]

    points = np.random.uniform(size=(100, 10))
    minvol = 1.0

    scale = 1. # np.random.uniform()
    for layer in AffineLayer, ScalingLayer:
        transformLayer = layer()
        transformLayer.optimize(points, points)
        for region_class in MLFriends, RobustEllipsoidRegion, SimpleRegion:
            region = region_class(points, transformLayer)
            r, f = region.compute_enlargement(minvol=minvol, nbootstraps=30)
            region.maxradiussq = r
            region.enlarge = f
            region.create_ellipsoid(minvol=minvol)

            for prop in proposals:
                print("test of proposal:", prop, "with region:", region_class, "layer:", layer)
                directions = prop(points, region, scale=scale)
                assert directions.shape == points.shape, (directions.shape, points.shape)
                #assert np.allclose(norms, scale), (norms, scale)


def test_direction_proposal_values():
    ndim = 10
    np.random.seed(12)
    region = make_region(ndim, nlive=400)
    ui = region.u[::2]
    
    scale = np.random.uniform()
    vcube = generate_cube_oriented_direction(ui, region, scale)
    assert vcube.shape == ui.shape
    assert vcube.sum(axis=1).shape == (len(ui),)
    assert ((vcube != 0).sum(axis=1) == 1).all(), vcube
    assert np.allclose(np.linalg.norm(vcube, axis=1), scale), (vcube, np.linalg.norm(vcube, axis=1), scale)

    vharm = generate_random_direction(ui, region, scale)
    assert (vharm != 0).all(), vharm
    vregionslice = generate_region_oriented_direction(ui, region, scale)
    assert (vregionslice != 0).all(), vregionslice
    vregionharm = generate_region_random_direction(ui, region, scale)
    assert (vregionharm != 0).all(), vregionharm
    vcubestd = generate_cube_oriented_direction_scaled(ui, region, scale)
    assert vcubestd.shape == ui.shape
    assert vcubestd.sum(axis=1).shape == (len(ui),)
    assert ((vcubestd != 0).sum(axis=1) == 1).all(), vcubestd


if __name__ == '__main__':
    test_stepsampler_cubegausswalk()
    test_direction_proposals()



# cython: language_level=3,fast_fail=True,warning_errors=True
# dython: language_level=3,annotate=True,profile=True,fast_fail=True,warning_errors=True
#!/usr/bin/env python
# coding: utf-8

import numpy as np
cimport numpy as np
from numpy import pi, nan as np_nan
cimport cython
from cython.parallel import prange

def submasks(mask, *masks):
    indices, = np.where(mask)
    for othermask in masks:
        indices = indices[othermask]
    return indices

@cython.boundscheck(False)
@cython.wraparound(False)
cdef _within_unit_cube(
    np.float_t [:, :] u, 
    np.uint8_t [:] acceptable, 
):
    cdef size_t popsize = u.shape[0]
    cdef size_t ndim = u.shape[1]
    cdef np.uint8_t good
    cdef np.int_t i, j

    for i in range(popsize):
        for j in range(ndim):
            if not 0.0 < u[i,j] < 1.0:
                acceptable[i] = 0
                break


def within_unit_cube(u):
    """whether all fields are between 0 and 1, for each row

    Parameters
    ----------
    u: np.array((npoints, ndim), dtype=float):
        points

    Returns
    ---------
    within: np.array(npoints, dtype=bool):
        for each point, whether it is within the unit cube
    """
    acceptable = np.ones(u.shape[0], dtype=bool)
    _within_unit_cube(u, acceptable)
    return acceptable

@cython.boundscheck(False)
@cython.wraparound(False)
cdef _evolve_prepare(
    np.ndarray[np.uint8_t, ndim=1] searching_left, 
    np.ndarray[np.uint8_t, ndim=1] searching_right,
    np.ndarray[np.uint8_t, ndim=1] search_right,
    np.ndarray[np.uint8_t, ndim=1] bisecting
):
    # define three mutually exclusive states: 
    # stepping out to the left, to the right, bisecting on the slice
    cdef size_t n = searching_left.shape[0]
    cdef np.int_t i
    for i in range(n):
        search_right[i] = not searching_left[i] and searching_right[i]
        bisecting[i] = not (searching_left[i] or searching_right[i])

def evolve_prepare(searching_left, searching_right):
    search_right = np.empty_like(searching_left)
    bisecting = np.empty_like(searching_left)
    _evolve_prepare(searching_left, searching_right, search_right, bisecting)
    return search_right, bisecting

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef evolve_update(
    np.ndarray[np.uint8_t, ndim=1] acceptable, 
    np.ndarray[np.float_t, ndim=1] Lnew, 
    np.float_t Lmin, 
    np.ndarray[np.uint8_t, ndim=1] search_right, 
    np.ndarray[np.uint8_t, ndim=1] bisecting, 
    np.float_t[:] currentt,
    np.float_t[:] current_left,
    np.float_t[:] current_right,
    np.uint8_t[:] searching_left,
    np.uint8_t[:] searching_right,
    np.uint8_t[:] success
):
    cdef size_t popsize = acceptable.shape[0]
    cdef np.int_t j = 0
    cdef np.int_t i
    cdef float my_nan = np_nan
    
    for i in range(popsize):
        if acceptable[i]:
            if Lnew[j] > Lmin:
                success[i] = 1
            j += 1

    for i in prange(popsize, nogil=True):
        # handle cases based on the result:
        # 1) step out further, if still accepting
        if success[i] != 0:
            if searching_left[i]:
                current_left[i] *= 2
            if search_right[i]:
                current_right[i] *= 2
        # 2) done stepping out, if rejected
        else:
            if searching_left[i]:
                searching_left[i] = 0
            if search_right[i]:
                searching_right[i] = 0
        # bisecting, rejected or not acceptable
        if bisecting[i]:
            if currentt[i] < 0:
                # bisect shrink left:
                current_left[i] = currentt[i]
            else:
                current_right[i] = currentt[i]
            # bisect accepted: start new slice and new generation there
            if success[i] != 0:
                currentt[i] = my_nan
        else:
            success[i] = 0

pnew_empty = np.empty((0,1))
Lnew_empty = np.empty(0)

def evolve(
    transform, loglike, Lmin, 
    currentu, currentL, currentt, currentv,
    current_left, current_right, searching_left, searching_right,
    log=False
):
    search_right, bisecting = evolve_prepare(searching_left, searching_right)

    unew = currentu
    unew[searching_left,:] = currentu[searching_left,:] + currentv[searching_left,:] * current_left[searching_left].reshape((-1,1))
    unew[search_right,:] = currentu[search_right,:] + currentv[search_right,:] * current_right[search_right].reshape((-1,1))
    currentt[bisecting] = np.random.uniform(current_left[bisecting], current_right[bisecting])
    unew[bisecting,:] = currentu[bisecting,:] + currentv[bisecting,:] * currentt[bisecting].reshape((-1,1))
        # assert np.isfinite(unew).all(), unew

    acceptable = within_unit_cube(unew)
    #acceptable = np.empty_like(searching_left)
    #_within_unit_cube(unew, acceptable)

    nc = 0
    if acceptable.any():
        pnew = transform(unew[acceptable,:])
        Lnew = loglike(pnew)
        nc += 1 # len(pnew)
    else:
        pnew = pnew_empty
        Lnew = Lnew_empty

    success = np.zeros_like(searching_left)
    evolve_update(
        acceptable, Lnew, Lmin, search_right, bisecting, currentt,
        current_left, current_right, searching_left, searching_right,
        success
    )

    return (
        (
        currentt, currentv,
        current_left, current_right, searching_left, searching_right), 
        (success, unew[success,:], pnew[success[acceptable],:], Lnew[success[acceptable]]), 
        nc
    )

cdef fill_directions(
    np.ndarray[np.float_t, ndim=2] v,
    np.ndarray[np.int_t, ndim=1] indices,
    float scale
):
    cdef size_t nsamples = v.shape[0]
    cdef size_t ndim = v.shape[0]
    cdef np.int_t i
    for i in range(nsamples):
        v[i, indices[i]] = scale

def generate_unit_directions(ui, region, scale=1):
    nsamples, ndim = ui.shape
    v = np.zeros((nsamples, ndim))
    # choose axis
    j = np.random.randint(ndim, size=nsamples)
    fill_directions(v, j, scale)
    #k = np.arange(nsamples)
    #v[k,j] = scale
    return v

def step_back(Lmin, allL, generation, currentt, log=False):
    # step back where step was excluded by Lmin increase
    # delete from the back until all are good:
    max_width = generation.max() + 1
    below_threshold = allL[:,:max_width] < Lmin
    problematic_parent = np.any(below_threshold, axis=1)
    if not problematic_parent.any():
        return
    parent_i, = np.where(problematic_parent)
    below_threshold_parent = below_threshold[parent_i,:]
    # first, all of them (because we already identified them)
    problematic = np.ones(len(parent_i), dtype=bool)
    #step = 0
    while True:
        if problematic.any():
            ii, = np.where(problematic)
            i = parent_i[problematic]
            g = generation[i]
            generation[i] -= 1
            currentt[i] = np_nan
            allL[i,g] = np_nan
            below_threshold_parent[problematic, g] = False
            #if log:
            #    print("resetting %d%%" % (problematic.meancount_good_generations() * 100), 'by', step, 'steps', 'to', g)
        else:
            break

        #step += 1
        problematic = np.any(below_threshold_parent, axis=1)



import numpy as np
import tempfile
import os
from ultranest.utils import vectorize, is_affine_transform, normalised_kendall_tau_distance, make_run_dir, verify_gradient
from ultranest.utils import distributed_work_chunk_size
from numpy.testing import assert_allclose
import pytest

def test_vectorize():
    
    def myfunc(x):
        return (x**2).sum()

    myvfunc = vectorize(myfunc)
    
    a = np.array([1.2, 2.3, 3.4])
    
    assert_allclose(np.array([myfunc(a)]), myvfunc([a]))
    b = np.array([[1.2, 2.3, 3.4], [1.2, 2.3, 3.4]])
    assert_allclose(np.array([myfunc(b[0]), myfunc(b[1])]), myvfunc(b))
    
    class FuncClass(object):
        def __call__(self, x):
            return (x**2).sum()
        def foo(self, x):
            return x
    
    mycaller = FuncClass()
    vectorize(mycaller)
    vectorize(mycaller.foo)


def test_is_affine_transform():
    na = 2**np.random.randint(1, 10)
    d = 2**np.random.randint(1, 3)
    a = np.random.uniform(-1, 1, size=(na, d))
    
    assert is_affine_transform(a, a)
    assert is_affine_transform(a, a * 2.0)
    assert is_affine_transform(a, a - 1)
    assert is_affine_transform(a, a * 10000 - 5000.)
    assert not is_affine_transform(a, a**2)

def test_tau():
    
    assert normalised_kendall_tau_distance(np.arange(400), np.arange(400)) == 0
    assert normalised_kendall_tau_distance(np.arange(2000), np.arange(2000)) == 0
    a = np.array([1, 2, 3, 4, 5])
    b = np.array([3, 4, 1, 2, 5])
    assert normalised_kendall_tau_distance(a, b) == 0.4
    i, j = np.meshgrid(np.arange(len(a)), np.arange(len(b)))
    assert normalised_kendall_tau_distance(a, b, i, j) == 0.4
    assert normalised_kendall_tau_distance(a, a, i, j) == 0
    
    try:
        normalised_kendall_tau_distance(np.arange(5), np.arange(10))
        raise Exception("expect error")
    except AssertionError:
        pass


def test_verify_gradient():
    ndim = 4
    sigma = 0.01
    sigma = np.logspace(-1, np.log10(sigma), ndim)
    width = 1 - 5 * sigma
    width[width < 1e-20] = 1e-20
    centers = (np.sin(np.arange(ndim)/2.) * width + 1.) / 2.

    def loglike(theta):
        return -0.5 * (((theta - centers)/sigma)**2).sum(axis=1) - 0.5 * np.log(2 * np.pi * sigma**2).sum()

    def transform(x):
        return x

    def transform_loglike_gradient(u):
        theta = u
        like = -0.5 * (((theta - centers)/sigma)**2).sum() - 0.5 * np.log(2 * np.pi * sigma**2).sum()
        grad = (theta - centers) / sigma
        return u, like, grad

    def gradient(theta):
        return (theta - centers) / sigma

    def wrong_gradient(theta):
        return -1000 * (theta - centers) / sigma

    verify_gradient(ndim, transform, loglike, transform_loglike_gradient, combination=True, verbose=True)
    verify_gradient(ndim, transform, loglike, gradient, verbose=True)
    failed = False
    try:
        verify_gradient(ndim, transform, loglike, wrong_gradient, verbose=True)
    except AssertionError:
        failed = True
    assert failed


def test_make_log_dirs():
    import shutil
    try:
        filepath = tempfile.mkdtemp()
        make_run_dir(filepath, max_run_num=3)
        assert os.path.exists(os.path.join(filepath, 'run1'))
        make_run_dir(filepath, max_run_num=3)
        assert os.path.exists(os.path.join(filepath, 'run2'))
        try:
            make_run_dir(filepath, max_run_num=3)
            assert False
        except ValueError:
            pass
    finally:
        shutil.rmtree(filepath)

@pytest.mark.parametrize("mpi_size", [1, 4, 10, 37, 53, 100, 1000, 513])
@pytest.mark.parametrize("num_live_points_missing", [0, 1, 4, 10, 17, 31, 100, 1000, 513])
def test_distributed_work_chunk_size(mpi_size, num_live_points_missing):
    processes = range(mpi_size)
    todo = [distributed_work_chunk_size(num_live_points_missing, rank, mpi_size) for rank in processes]
    assert sum(todo) == num_live_points_missing
    assert max(todo) - min(todo) in {0, 1}



"""
Vectorized step samplers
------------------------

Likelihood based on GPUs (model emulators based on neural networks,
or JAX implementations) can evaluate hundreds of points as efficiently
as one point. The implementations in this module leverage this power,
by providing random walks of populations of walkers.
"""

import numpy as np
from ultranest.utils import submasks
from ultranest.stepfuncs import evolve, step_back
from ultranest.stepfuncs import generate_cube_oriented_direction, generate_cube_oriented_direction_scaled
from ultranest.stepfuncs import generate_random_direction, generate_region_oriented_direction, generate_region_random_direction
from ultranest.stepfuncs import generate_differential_direction, generate_mixture_random_direction
import scipy.stats


def unitcube_line_intersection(ray_origin, ray_direction):
    r"""Compute intersection of a line (ray) and a unit box (0:1 in all axes).

    Based on
    http://www.iquilezles.org/www/articles/intersectors/intersectors.htm

    Parameters
    -----------
    ray_origin: array of vectors
        starting point of line
    ray_direction: vector
        line direction vector

    Returns
    --------
    tleft: array
        negative intersection point distance from ray\_origin in units in ray\_direction
    tright: array
        positive intersection point distance from ray\_origin in units in ray\_direction

    """
    # make sure ray starts inside the box
    assert (ray_origin >= 0).all(), ray_origin
    assert (ray_origin <= 1).all(), ray_origin
    assert ((ray_direction**2).sum()**0.5 > 1e-200).all(), ray_direction

    # step size
    with np.errstate(divide='ignore', invalid='ignore'):
        m = 1. / ray_direction
        n = m * (ray_origin - 0.5)
        k = np.abs(m) * 0.5
        # line coordinates of intersection
        # find first intersecting coordinate
        t1 = -n - k
        t2 = -n + k
        return np.nanmax(t1, axis=1), np.nanmin(t2, axis=1)

def diagnose_move_distances(region, ustart, ufinal):
    """Compares random walk travel distance to MLFriends radius.

    Compares in whitened space (t-space), the L2 norm between final
    point and starting point to the MLFriends bootstrapped radius.

    Parameters
    ----------
    region: MLFriends
        built region
    ustart: array
        starting positions
    ufinal: array
        final positions

    Returns
    -------
    far_enough: bool
        whether the distance is larger than the radius
    move_distance: float
        distance between start and final point in whitened space
    reference_distance: float
        MLFriends radius
    """
    assert ustart.shape == ufinal.shape, (ustart.shape, ufinal.shape)
    tstart = region.transformLayer.transform(ustart)
    tfinal = region.transformLayer.transform(ufinal)
    d2 = ((tstart - tfinal)**2).sum(axis=1)
    far_enough = d2 > region.maxradiussq

    return far_enough, [d2**0.5, region.maxradiussq**0.5]

class GenericPopulationSampler():
    def plot(self, filename):
        """Plot sampler statistics.

        Parameters
        -----------
        filename: str
            Stores plot into ``filename`` and data into
            ``filename + ".txt.gz"``.
        """
        if len(self.logstat) == 0:
            return

        import matplotlib.pyplot as plt
        plt.figure(figsize=(10, 1 + 3 * len(self.logstat_labels)))
        for i, label in enumerate(self.logstat_labels):
            part = [entry[i] for entry in self.logstat]
            plt.subplot(len(self.logstat_labels), 1, 1 + i)
            plt.ylabel(label)
            plt.plot(part)
            x = []
            y = []
            for j in range(0, len(part), 20):
                x.append(j)
                y.append(np.mean(part[j:j + 20]))
            plt.plot(x, y)
            if np.min(part) > 0:
                plt.yscale('log')
        plt.savefig(filename, bbox_inches='tight')
        np.savetxt(filename + '.txt.gz', self.logstat,
                   header=','.join(self.logstat_labels), delimiter=',')
        plt.close()

    @property
    def mean_jump_distance(self):
        """Geometric mean jump distance."""
        if len(self.logstat) == 0:
            return np.nan
        return np.exp(np.nanmean(np.log([entry[-1] for entry in self.logstat])))

    @property
    def far_enough_fraction(self):
        """Fraction of jumps exceeding reference distance."""
        if len(self.logstat) == 0:
            return np.nan
        return np.nanmean([entry[-2] for entry in self.logstat])

    def get_info_dict(self):
        return dict(
            num_logs=len(self.logstat),
            rejection_rate=np.nanmean([entry[0] for entry in self.logstat]),
            mean_scale=np.nanmean([entry[1] for entry in self.logstat]),
            mean_nsteps=np.nanmean([entry[2] for entry in self.logstat]),
            mean_distance=self.mean_jump_distance,
            frac_far_enough=self.far_enough_fraction,
            last_logstat=dict(zip(self.logstat_labels, self.logstat[-1] if len(self.logstat) > 1 else [np.nan] * len(self.logstat_labels)))
        )


    def print_diagnostic(self):
        """Print diagnostic of step sampler performance."""
        if len(self.logstat) == 0:
            print("diagnostic unavailable, no recorded steps found")
            return
        frac_farenough = self.far_enough_fraction
        average_distance = self.mean_jump_distance
        if frac_farenough < 0.5:
            advice = ': very fishy. Double nsteps and see if fraction and lnZ change)'
        elif frac_farenough < 0.66:
            advice = ': fishy. Double nsteps and see if fraction and lnZ change)'
        else:
            advice = ' (should be >50%)'
        print('step sampler diagnostic: jump distance %.2f (should be >1), far enough fraction: %.2f%% %s' % (
            average_distance, frac_farenough * 100, advice))

    def plot_jump_diagnostic_histogram(self, filename, **kwargs):
        """Plot jump diagnostic histogram."""
        if len(self.logstat) == 0:
            return
        import matplotlib.pyplot as plt
        plt.hist(np.log10([entry[-1] for entry in self.logstat]), **kwargs)
        ylo, yhi = plt.ylim()
        plt.vlines(self.mean_jump_distance, ylo, yhi)
        plt.ylim(ylo, yhi)
        plt.xlabel('log(relative step distance)')
        plt.ylabel('Frequency')
        plt.savefig(filename, bbox_inches='tight')
        plt.close()
    

class PopulationRandomWalkSampler(GenericPopulationSampler):
    """Vectorized Gaussian Random Walk sampler."""

    def __init__(
        self, popsize, nsteps, generate_direction, scale,
        scale_adapt_factor=0.9, scale_min=1e-20, scale_max=20, log=False, logfile=None
    ):
        """Initialise.

        Parameters
        ----------
        popsize: int
            number of walkers to maintain.
            this should be fairly large (~100), if too large you probably get memory issues
            Also, some results have to be discarded as the likelihood threshold increases.
            Observe the nested sampling efficiency.
        nsteps: int
            number of steps to take until the found point is accepted as independent.
            To find the right value, see :py:class:`ultranest.calibrator.ReactiveNestedCalibrator`
        generate_direction: function
            Function that gives proposal kernel shape, one of:
            :py:func:`ultranest.popstepsampler.generate_cube_oriented_direction`
            :py:func:`ultranest.popstepsampler.generate_cube_oriented_direction_scaled`
            :py:func:`ultranest.popstepsampler.generate_random_direction`
            :py:func:`ultranest.popstepsampler.generate_region_oriented_direction`
            :py:func:`ultranest.popstepsampler.generate_region_random_direction`
        scale: float
            initial guess for the proposal scaling factor
        scale_adapt_factor: float
            if 1, no adapting is done.
            if <1, the scale is increased if the acceptance rate is below 23.4%,
            or decreased if it is above, by *scale_adapt_factor*.
        scale_min: float
            lowest value allowed for scale, do not adapt down further
        scale_max: float
            highest value allowed for scale, do not adapt up further
        logfile: file
            where to print the current scaling factor and acceptance rate

        """
        self.nsteps = nsteps
        self.nrejects = 0
        self.scale = scale
        self.ncalls = 0
        assert scale_adapt_factor <= 1
        self.scale_adapt_factor = scale_adapt_factor
        self.scale_min = scale_min
        self.scale_max = scale_max

        self.log = log
        self.logfile = logfile
        self.logstat = []
        self.logstat_labels = ['accept_rate', 'efficiency', 'scale', 'far_enough', 'mean_rel_jump']
        self.prepared_samples = []

        self.popsize = popsize
        self.generate_direction = generate_direction

    def __str__(self):
        """Return string representation."""
        return 'PopulationRandomWalkSampler(popsize=%d, nsteps=%d, generate_direction=%s, scale=%.g)' % (
            self.popsize, self.nsteps, self.generate_direction, self.scale)

    def region_changed(self, Ls, region):
        """Act upon region changed. Currently unused."""
        pass

    def __next__(
        self, region, Lmin, us, Ls, transform, loglike, ndraw=10,
        plot=False, tregion=None, log=False
    ):
        """Sample a new live point.

        Parameters
        ----------
        region: MLFriends object
            Region
        Lmin: float
            current log-likelihood threshold
        us: np.array((nlive, ndim))
            live points
        Ls: np.array(nlive)
            loglikelihoods live points
        transform: function
            prior transform function
        loglike: function
            loglikelihood function
        ndraw: int
            not used
        plot: bool
            not used
        tregion: bool
            not used
        log: bool
            not used

        Returns
        -------
        u: np.array(ndim) or None
            new point coordinates (None if not yet available)
        p: np.array(nparams) or None
            new point transformed coordinates (None if not yet available)
        L: float or None
            new point likelihood (None if not yet available)
        nc: int

        """
        nlive, ndim = us.shape

        # fill if empty:
        if len(self.prepared_samples) == 0:
            # choose live points
            ilive = np.random.randint(0, nlive, size=self.popsize)
            allu = us[ilive,:]
            allp = None
            allL = Ls[ilive]
            nc = self.nsteps * self.popsize
            nrejects_expected = self.nrejects + self.nsteps * self.popsize * (1 - 0.234)

            for i in range(self.nsteps):
                # perturb walker population
                v = self.generate_direction(allu, region, self.scale)
                # compute intersection of u + t * v with unit cube
                tleft, tright = unitcube_line_intersection(allu, v)
                proposed_t = scipy.stats.truncnorm.rvs(tleft, tright, loc=0, scale=1).reshape((-1, 1))

                proposed_u = allu + v * proposed_t
                mask_outside = ~np.logical_and(proposed_u > 0, proposed_u < 1).all(axis=1)
                assert not mask_outside.any(), proposed_u[mask_outside, :]

                proposed_p = transform(proposed_u)
                # accept if likelihood threshold exceeded
                proposed_L = loglike(proposed_p)
                mask_accept = proposed_L > Lmin
                self.nrejects += (~mask_accept).sum()
                allu[mask_accept,:] = proposed_u[mask_accept,:]
                if allp is None:
                    del allp
                    allp = proposed_p * np.nan
                allp[mask_accept,:] = proposed_p[mask_accept,:]
                allL[mask_accept] = proposed_L[mask_accept]
            assert np.isfinite(allp).all(), 'some walkers never moved! Double nsteps of PopulationRandomWalkSampler.'
            far_enough, (move_distance, reference_distance) = diagnose_move_distances(region, us[ilive[mask_accept],:], allu[mask_accept,:])
            self.prepared_samples = list(zip(allu, allp, allL))

            self.logstat.append([
                mask_accept.mean(),
                1 - (self.nrejects - (nrejects_expected - self.nsteps * self.popsize * (1 - 0.234))) / (self.nsteps * self.popsize),
                self.scale,
                self.nsteps,
                np.mean(far_enough),
                np.exp(np.mean(np.log(move_distance / reference_distance + 1e-10)))
            ])
            if self.logfile:
                self.logfile.write("rescale\t%.4f\t%.4f\t%g\t%.4f%g\n" % self.logstat[-1])

            # adapt slightly
            if self.nrejects > nrejects_expected and self.scale > self.scale_min:
                # lots of rejects, decrease scale
                self.scale *= self.scale_adapt_factor
            elif self.nrejects < nrejects_expected and self.scale < self.scale_max:
                self.scale /= self.scale_adapt_factor
        else:
            nc = 0

        u, p, L = self.prepared_samples.pop(0)
        return u, p, L, nc


class PopulationSliceSampler(GenericPopulationSampler):
    """Vectorized slice/HARM sampler.

    Can revert until all previous steps have likelihoods allL above Lmin.
    Updates currentt, generation and allL, in-place.
    """

    def __init__(
        self, popsize, nsteps, generate_direction, scale=1.0,
        scale_adapt_factor=0.9, log=False, logfile=None
    ):
        """Initialise.

        Parameters
        ----------
        popsize: int
            number of walkers to maintain
        nsteps: int
            number of steps to take until the found point is accepted as independent.
            To find the right value, see :py:class:`ultranest.calibrator.ReactiveNestedCalibrator`
        generate_direction: function `(u, region, scale) -> v`
            function such as `generate_unit_directions`, which
            generates a random slice direction.
        scale: float
            initial guess scale for the length of the slice
        scale_adapt_factor: float
            smoothing factor for updating scale.
            if near 1, scale is barely updating, if near 0,
            the last slice length is used as a initial guess for the next.

        """
        self.nsteps = nsteps
        self.nrejects = 0
        self.scale = scale
        self.scale_adapt_factor = scale_adapt_factor
        self.allu = []
        self.allL = []
        self.currentt = []
        self.currentv = []
        self.currentp = []
        self.generation = []
        self.current_left = []
        self.current_right = []
        self.searching_left = []
        self.searching_right = []
        self.ringindex = 0

        self.log = log
        self.logfile = logfile
        self.logstat = []
        self.logstat_labels = ['accept_rate', 'efficiency', 'scale', 'far_enough', 'mean_rel_jump']

        self.popsize = popsize
        self.generate_direction = generate_direction

    def __str__(self):
        """Return string representation."""
        return 'PopulationSliceSampler(popsize=%d, nsteps=%d, generate_direction=%s, scale=%.g)' % (
            self.popsize, self.nsteps, self.generate_direction, self.scale)

    def region_changed(self, Ls, region):
        """Act upon region changed. Currently unused."""
        # self.scale = region.us.std(axis=1).mean()
        if self.logfile:
            self.logfile.write("region-update\t%g\t%g\n" % (self.scale, region.us.std(axis=1).mean()))

    def _setup(self, ndim):
        """Allocate arrays."""
        self.allu = np.zeros((self.popsize, self.nsteps + 1, ndim)) + np.nan
        self.allL = np.zeros((self.popsize, self.nsteps + 1)) + np.nan
        self.currentt = np.zeros(self.popsize) + np.nan
        self.currentv = np.zeros((self.popsize, ndim)) + np.nan
        self.generation = np.zeros(self.popsize, dtype=int) - 1
        self.current_left = np.zeros(self.popsize)
        self.current_right = np.zeros(self.popsize)
        self.searching_left = np.zeros(self.popsize, dtype=bool)
        self.searching_right = np.zeros(self.popsize, dtype=bool)

    def setup_start(self, us, Ls, starting):
        """Initialize walker starting points.

        For iteration zero, randomly selects a live point as starting point.

        Parameters
        ----------
        us: np.array((nlive, ndim))
            live points
        Ls: np.array(nlive)
            loglikelihoods live points
        starting: np.array(nwalkers, dtype=bool)
            which walkers to initialize.

        """
        if self.log:
            print("setting up:", starting)
        nlive = len(us)
        i = np.random.randint(nlive, size=starting.sum())

        if not starting.all():
            while starting[self.ringindex]:
                # if the one we are waiting for is being restarted,
                # we may as well pick the next one to wait for
                # because every other one is started from a random point
                # as well
                self.shift()

        self.allu[starting,0] = us[i]
        self.allL[starting,0] = Ls[i]
        self.generation[starting] = 0

    @property
    def status(self):
        """Return compact string representation of the current status."""
        s1 = ('G:' + ''.join(['%d' % g if g >= 0 else '_' for g in self.generation]))
        s2 = ('S:' + ''.join([
            'S' if not np.isfinite(self.currentt[i]) else 'L' if self.searching_left[i] else 'R' if self.searching_right[i] else 'B'
            for i in range(self.popsize)]))
        return s1 + '  ' + s2

    def setup_brackets(self, mask_starting, region):
        """Pick starting direction and range for slice.

        Parameters
        ----------
        region: MLFriends object
            Region
        mask_starting: np.array(nwalkers, dtype=bool)
            which walkers to set up.

        """
        if self.log:
            print("starting brackets:", mask_starting)
        i_starting, = np.where(mask_starting)
        self.current_left[i_starting] = -self.scale
        self.current_right[i_starting] = self.scale
        self.searching_left[i_starting] = True
        self.searching_right[i_starting] = True
        self.currentt[i_starting] = 0
        # choose direction for new slice
        self.currentv[i_starting,:] = self.generate_direction(
            self.allu[i_starting, self.generation[i_starting]],
            region)

    def _setup_currentp(self, nparams):
        if self.log:
            print("setting currentp")
        self.currentp = np.zeros((self.popsize, nparams)) + np.nan

    def advance(self, transform, loglike, Lmin, region):
        """Advance the walker population.

        Parameters
        ----------
        transform: function
            prior transform function
        loglike: function
            loglikelihood function
        Lmin: float
            current log-likelihood threshold

        """
        movable = self.generation < self.nsteps
        all_movable = movable.all()
        # print("moving ", movable.sum(), self.popsize)
        if all_movable:
            i = np.arange(self.popsize)
            args = [
                self.allu[i, self.generation],
                self.allL[i, self.generation],
                # pass values directly
                self.currentt,
                self.currentv,
                self.current_left,
                self.current_right,
                self.searching_left,
                self.searching_right
            ]
            del i
        else:
            args = [
                self.allu[movable, self.generation[movable]],
                self.allL[movable, self.generation[movable]],
                # this makes copies
                self.currentt[movable],
                self.currentv[movable],
                self.current_left[movable],
                self.current_right[movable],
                self.searching_left[movable],
                self.searching_right[movable]
            ]
        if self.log:
            print("evolve will advance:", movable)

        (
            (
                currentt, currentv,
                current_left, current_right, searching_left, searching_right
            ),
            (success, unew, pnew, Lnew),
            nc
        ) = evolve(transform, loglike, Lmin, *args)
        
        far_enough, (move_distance, reference_distance) = diagnose_move_distances(region, args[0][success,:], unew)
        self.logstat.append([
            success.mean(),
            self.scale,
            self.nsteps,
            np.mean(far_enough),
            np.exp(np.mean(np.log(move_distance / reference_distance + 1e-10)))
        ])
        if self.logfile:
            self.logfile.write("rescale\t%.4f\t%.4f\t%g\t%.4f%g\n" % self.logstat[-1])

        if self.log:
            print("movable", movable.shape, movable.sum(), success.shape)
        moved = submasks(movable, success)
        if self.log:
            print("evolve moved:", moved)
        self.generation[moved] += 1
        if len(pnew) > 0:
            if len(self.currentp) == 0:
                self._setup_currentp(nparams=pnew.shape[1])

            if self.log:
                print("currentp", self.currentp[moved,:].shape, pnew.shape)
            self.currentp[moved,:] = pnew

        # update with what we learned
        # print(currentu.shape, currentL.shape, success.shape, self.generation[movable])
        self.allu[moved, self.generation[moved]] = unew
        self.allL[moved, self.generation[moved]] = Lnew
        if all_movable:
            # in this case, the values were directly overwritten
            pass
        else:
            self.currentt[movable] = currentt
            self.currentv[movable] = currentv
            self.current_left[movable] = current_left
            self.current_right[movable] = current_right
            self.searching_left[movable] = searching_left
            self.searching_right[movable] = searching_right
        return nc

    def shift(self):
        """Update walker from which to pick next."""
        # this is a ring buffer
        # shift index forward, wrapping around
        # this is better than copying memory around when a element is removed
        self.ringindex = (self.ringindex + 1) % self.popsize

    def __next__(
        self, region, Lmin, us, Ls, transform, loglike, ndraw=10,
        plot=False, tregion=None, log=False
    ):
        """Sample a new live point.

        Parameters
        ----------
        region: MLFriends object
            Region
        Lmin: float
            current log-likelihood threshold
        us: np.array((nlive, ndim))
            live points
        Ls: np.array(nlive)
            loglikelihoods live points
        transform: function
            prior transform function
        loglike: function
            loglikelihood function
        ndraw: int
            not used
        plot: bool
            not used
        tregion: bool
            not used
        log: bool
            not used

        Returns
        -------
        u: np.array(ndim) or None
            new point coordinates (None if not yet available)
        p: np.array(nparams) or None
            new point transformed coordinates (None if not yet available)
        L: float or None
            new point likelihood (None if not yet available)
        nc: int

        """
        nlive, ndim = us.shape
        # initialize
        if len(self.allu) == 0:
            self._setup(ndim)

        step_back(Lmin, self.allL, self.generation, self.currentt)

        starting = self.generation < 0
        if starting.any():
            self.setup_start(us[Ls > Lmin], Ls[Ls > Lmin], starting)
        assert (self.generation >= 0).all(), self.generation

        # find those where bracket is undefined:
        mask_starting = ~np.isfinite(self.currentt)
        if mask_starting.any():
            self.setup_brackets(mask_starting, region)

        if self.log:
            print(str(self), "(before)")
        nc = self.advance(transform, loglike, Lmin, region)
        if self.log:
            print(str(self), "(after)")

        # harvest top individual if possible
        if self.generation[self.ringindex] == self.nsteps:
            if self.log:
                print("have a candidate")
            u, p, L = self.allu[self.ringindex, self.nsteps, :].copy(), self.currentp[self.ringindex, :].copy(), self.allL[self.ringindex, self.nsteps].copy()
            assert np.isfinite(u).all(), u
            assert np.isfinite(p).all(), p
            self.generation[self.ringindex] = -1
            self.currentt[self.ringindex] = np.nan
            self.allu[self.ringindex,:,:] = np.nan
            self.allL[self.ringindex,:] = np.nan

            # adjust guess length
            newscale = (self.current_right[self.ringindex] - self.current_left[self.ringindex]) / 2
            self.scale = self.scale * 0.9 + 0.1 * newscale

            self.shift()
            return u, p, L, nc
        else:
            return None, None, None, nc


__all__ = [
    "generate_cube_oriented_direction", "generate_cube_oriented_direction_scaled",
    "generate_random_direction", "generate_region_oriented_direction", "generate_region_random_direction",
    "PopulationRandomWalkSampler", "PopulationSliceSampler"]



"""Sparsely sampled, virtual sampling path.

Supports reflections at unit cube boundaries, and regions.
"""


import numpy as np
from numpy.linalg import norm
import matplotlib.pyplot as plt


def nearest_box_intersection_line(ray_origin, ray_direction, fwd=True):
    r"""Compute intersection of a line (ray) and a unit box (0:1 in all axes).

    Based on
    http://www.iquilezles.org/www/articles/intersectors/intersectors.htm

    To continue forward traversing at the reflection point use::

        while True:
            # update current point x
            x, _, i = box_line_intersection(x, v)
            # change direction
            v[i] *= -1

    Parameters
    -----------
    ray_origin: vector
        starting point of line
    ray_direction: vector
        line direction vector

    Returns
    --------
    p: vector
        intersection point
    t: float
        intersection point distance from ray\_origin in units in ray\_direction
    i: int
        axes which change direction at pN

    """
    # make sure ray starts inside the box
    assert (ray_origin >= 0).all(), ray_origin
    assert (ray_origin <= 1).all(), ray_origin
    assert ((ray_direction**2).sum()**0.5 > 1e-200).all(), ray_direction

    # step size
    with np.errstate(divide='ignore', invalid='ignore'):
        m = 1. / ray_direction
        n = m * (ray_origin - 0.5)
        k = np.abs(m) * 0.5
        # line coordinates of intersection
        # find first intersecting coordinate
        if fwd:
            t2 = -n + k
            tF = np.nanmin(t2)
            iF = np.where(t2 == tF)[0]
        else:
            t1 = -n - k
            tF = np.nanmax(t1)
            iF = np.where(t1 == tF)[0]

    pF = ray_origin + ray_direction * tF
    eps = 1e-6
    assert (pF >= -eps).all(), (pF, ray_origin, ray_direction)
    assert (pF <= 1 + eps).all(), (pF, ray_origin, ray_direction)
    pF[pF < 0] = 0
    pF[pF > 1] = 1
    return pF, tF, iF


def box_line_intersection(ray_origin, ray_direction):
    """Find intersections of a line with the unit cube, in both sides.

    Parameters
    -----------
    ray_origin: vector
        starting point of line
    ray_direction: vector
        line direction vector

    Returns
    --------
    left: nearest_box_intersection_line return value
        from negative direction
    right: nearest_box_intersection_line return value
        from positive direction

    """
    pF, tF, iF = nearest_box_intersection_line(ray_origin, ray_direction, fwd=True)
    pN, tN, iN = nearest_box_intersection_line(ray_origin, ray_direction, fwd=False)
    if tN > tF or tF < 0:
        assert False, "no intersection"
    return (pN, tN, iN), (pF, tF, iF)


def linear_steps_with_reflection(ray_origin, ray_direction, t, wrapped_dims=None):
    """Go `t` steps in direction `ray_direction` from `ray_origin`.

    Reflect off the unit cube if encountered, respecting wrapped dimensions.
    In any case, the distance should be ``t * ray_direction``.

    Returns
    --------
    new_point: vector
        end point
    new_direction: vector
        new direction.

    """
    if t == 0:
        return ray_origin, ray_direction
    if t < 0:
        new_point, new_direction = linear_steps_with_reflection(ray_origin, -ray_direction, -t)
        return new_point, -new_direction

    if wrapped_dims is not None:
        reflected = np.zeros(len(ray_origin), dtype=bool)

    tleft = 1.0 * t
    while True:
        p, t, i = nearest_box_intersection_line(ray_origin, ray_direction, fwd=True)
        # print(p, t, i, ray_origin, ray_direction)
        assert np.isfinite(p).all()
        assert t >= 0, t
        if tleft <= t:  # stopping before reaching any border
            assert np.all(ray_origin + tleft * ray_direction >= 0), (ray_origin, tleft, ray_direction)
            assert np.all(ray_origin + tleft * ray_direction <= 1), (ray_origin, tleft, ray_direction)
            return ray_origin + tleft * ray_direction, ray_direction
        # go to reflection point
        ray_origin = p
        assert np.isfinite(ray_origin).all(), ray_origin
        # reflect
        ray_direction = ray_direction.copy()
        if wrapped_dims is None:
            ray_direction[i] *= -1
        else:
            # if we already once bumped into that (wrapped) axis,
            # do not continue but return this as end point
            if np.logical_and(reflected[i], wrapped_dims[i]).any():
                return ray_origin, ray_direction

            # note which axes we already flipped
            reflected[i] = True

            # in wrapped axes, we can keep going. Otherwise, reflects
            ray_direction[i] *= np.where(wrapped_dims[i], 1, -1)

            # in the i axes, we should wrap the coordinates
            assert np.logical_or(np.isclose(ray_origin[i], 1), np.isclose(ray_origin[i], 0)).all(), ray_origin[i]
            ray_origin[i] = np.where(wrapped_dims[i], 1 - ray_origin[i], ray_origin[i])

        assert np.isfinite(ray_direction).all(), ray_direction
        # reduce remaining distance
        tleft -= t


def get_sphere_tangent(sphere_center, edge_point):
    """Compute tangent at sphere surface point.

    Assume a sphere centered at sphere_center with radius
    so that edge_point is on the surface. At edge_point, in
    which direction does the normal vector point?

    Parameters
    -----------
    sphere_center: vector
        center of sphere
    edge_point: vector
        point at the surface

    Returns
    --------
    tangent: vector
        vector pointing to the sphere center.

    """
    arrow = sphere_center - edge_point
    return arrow / norm(arrow)


def get_sphere_tangents(sphere_center, edge_point):
    """Compute tangent at sphere surface point.

    Assume a sphere centered at sphere_center with radius
    so that edge_point is on the surface. At edge_point, in
    which direction does the normal vector point?

    This function is vectorized and handles arrays of arguments.

    Parameters
    -----------
    sphere_center: array
        centers of spheres
    edge_point: array
        points at the surface

    Returns
    --------
    tangent: array
        vectors pointing to the sphere center.

    """
    arrow = sphere_center - edge_point
    return arrow / norm(arrow, axis=1).reshape((-1, 1))


def reflect(v, normal):
    """Reflect vector ``v`` off a ``normal`` vector, return new direction vector."""
    return v - 2 * (normal * v).sum() * normal


def distances(direction, center, r=1):
    """Compute sphere-line intersection.

    Parameters
    -----------
    direction: vector
        direction vector (line starts at 0)
    center: vector
        center of sphere (coordinate vector)
    r: float
        radius of sphere

    Returns
    --------
    tpos, tneg: floats
        the positive and negative coordinate along the `l` vector where `r` is intersected.
        If no intersection, throws AssertError.

    """
    loc = (direction * center).sum()
    osqrnorm = (center**2).sum()
    # print(loc.shape, loc.shape, osqrnorm.shape)
    rootterm = loc**2 - osqrnorm + r**2
    # make sure we are crossing the sphere
    assert (rootterm > 0).all(), rootterm
    return -loc + rootterm**0.5, -loc - rootterm**0.5


def isunitlength(vec):
    """Verify that `vec` is of unit length."""
    assert np.isclose(norm(vec), 1), norm(vec)


def angle(a, b):
    """Compute dot product between vectors `a` and `b`.

    The arccos of the return value would give an actual angle.
    """
    # anorm = (a**2).sum()**0.5
    # bnorm = (b**2).sum()**0.5
    return (a * b).sum()  # / anorm / bnorm


def extrapolate_ahead(dj, xj, vj, contourpath=None):
    """Make `di` steps of size `vj` from `xj`.

    Reflect off unit cube if necessary.
    """
    assert dj == int(dj)

    # optimistically try to go there directly

    xk, vk = linear_steps_with_reflection(xj, vj, dj)

    return xk, vk  # deactivate feature below

    if contourpath is None:
        return xk, vk

    # check if we can already tell that the point will be outside
    region = contourpath.region
    if contourpath.region.inside(xk.reshape((1, -1))):
        return xk, vk

    # find first point outside region
    sign = 1 if dj > 0 else -1
    d = np.arange(0, dj, sign)
    first_point_outside = dj, xk, vk
    for di in d:
        xi, vi = linear_steps_with_reflection(xj, vj, di)
        if not region.inside(xk.reshape((1, -1))):
            first_point_outside = di, xi, vi
            break

    # reflect at this point (first outside)
    dout, reflpoint, v = first_point_outside

    if dout == 0:
        # already the starting point is outside.
        # return extrapolation and hope caller handles it
        return xk, vk

    # reversing:
    normal = contourpath.gradient(reflpoint)  # , v * sign)
    if normal is None:
        vnew = -v
    else:
        vnew = (v - 2 * angle(normal, v) * normal) * sign
    assert vnew.shape == v.shape, (vnew.shape, v.shape)
    assert np.isclose(norm(vnew), norm(v)), (vnew, v, norm(vnew), norm(v))

    # make one step (xl replaces first_point_outside/reflpoint)
    xl, vl = linear_steps_with_reflection(reflpoint, vnew, sign)
    # how many steps are still to do?
    dleft = dj - dout

    # make that many step in that direction, by recursing.

    # it is possible that this point is also outside. The next iteration
    # will catch that case.

    return extrapolate_ahead(dleft, xl, vl, contourpath=contourpath)


def interpolate(i, points, fwd_possible, rwd_possible, contourpath=None):
    """Interpolate a point on the path indicated by `points`.

    Given a sparsely sampled track (stored in .points),
    potentially encountering reflections,
    extract the corrdinates of the point with index `i`.
    That point may not have been evaluated yet.

    Parameters
    -----------
    i: int
        position on track to return.
    points: list of tuples (index, coordinate, direction, loglike)
        points on the path
    fwd_possible: bool
        whether the path could be extended in the positive direction.
    rwd_possible: bool
        whether the path could be extended in the negative direction.
    contourpath: ContourPath
        Use region to reflect. Not used at the moment.

    """
    points_before = [(j, xj, vj, Lj) for j, xj, vj, Lj in points if j <= i]
    points_after = [(j, xj, vj, Lj) for j, xj, vj, Lj in points if j >= i]

    # check if the point after is really after i
    if len(points_after) == 0 and not fwd_possible:
        # the path cannot continue, and i does not exist.
        # print("    interpolate_point %d: the path cannot continue fwd, and i does not exist." % i)
        j, xj, vj, Lj = max(points_before)
        return xj, vj, Lj, False

    # check if the point before is really before i
    if len(points_before) == 0 and not rwd_possible:
        # the path cannot continue, and i does not exist.
        k, xk, vk, Lk = min(points_after)
        # print("    interpolate_point %d: the path cannot continue rwd, and i does not exist." % i)
        return xk, vk, Lk, False

    if len(points_before) == 0 or len(points_after) == 0:
        # return None, None, None, False
        raise KeyError("cannot extrapolate outside path")

    j, xj, vj, Lj = max(points_before)
    k, xk, vk, Lk = min(points_after)

    # print("    interpolate_point %d between %d-%d" % (i, j, k))
    if j == i:  # we have this exact point in the chain
        return xj, vj, Lj, True

    assert not k == i  # otherwise the above would be true too

    # expand_to_step explores each reflection in detail, so
    # any points with change in v should have j == i
    # therefore we can assume:
    # assert (vj == vk).all()
    # this ^ is not true, because reflections on the unit cube can
    # occur, and change v without requiring a intermediate point.

    # j....i....k
    xl1, vj1 = extrapolate_ahead(i - j, xj, vj, contourpath=contourpath)
    xl2, vj2 = extrapolate_ahead(i - k, xk, vk, contourpath=contourpath)
    assert np.allclose(xl1, xl2), (xl1, xl2, i, j, k, xj, vj, xk, vk)
    assert np.allclose(vj1, vj2), (xl1, vj1, xl2, vj2, i, j, k, xj, vj, xk, vk)
    xl = xl1

    # xl = interpolate_between_two_points(i, xj, j, xk, k)
    # the new point is then just a linear interpolation
    # w = (i - k) * 1. / (j - k)
    # xl = xj * w + (1 - w) * xk

    return xl, vj, None, True


class SamplingPath(object):
    """Path described by a (potentially sparse) sequence of points.

    Convention of the stored point tuple ``(i, x, v, L)``:
    `i`: index (0 is starting point)
    `x`: point
    `v`: direction
    `L`: loglikelihood value
    """

    def __init__(self, x0, v0, L0):
        """Initialise with path starting point.

        Starting point (`x0`), direction (`v0`) and
        loglikelihood value (`L0`) of the path. Is given index 0.
        """
        self.reset(x0, v0, L0)

    def add(self, i, xi, vi, Li):
        """Add point `xi`, direction `vi` and value `Li` with index `i` to the path."""
        assert Li is not None
        assert len(xi.shape) == 1, (xi, xi.shape)
        assert len(vi.shape) == 1, (vi, vi.shape)
        assert len(np.shape(Li)) == 0, (Li, Li.shape)
        self.points.append((i, xi, vi, Li))

    def reset(self, x0, v0, L0):
        """Reset path, start from ``x0, v0, L0``."""
        self.points = []
        self.add(0, x0, v0, L0)
        self.fwd_possible = True
        self.rwd_possible = True

    def plot(self, **kwargs):
        """Plot the current path.

        Only uses first two dimensions.
        """
        x = np.array([x for i, x, v, L in sorted(self.points)])
        p, = plt.plot(x[:,0], x[:,1], 'o ', **kwargs)
        ilo, _, _, _ = min(self.points)
        ihi, _, _, _ = max(self.points)
        x = np.array([self.interpolate(i)[0] for i in range(ilo, ihi + 1)])
        kwargs['color'] = p.get_color()
        plt.plot(x[:,0], x[:,1], 'o-', ms=4, mfc='None', **kwargs)

    def interpolate(self, i):
        """Interpolate point with index `i` on path."""
        return interpolate(i, self.points, fwd_possible=self.fwd_possible, rwd_possible=self.rwd_possible)

    def extrapolate(self, i):
        """Advance beyond the current path, extrapolate from the end point.

        Parameters
        -----------
        i: int
            index on path.

        returns
        --------
        coords: vector
            coordinates of the new point.

        """
        if i >= 0:
            j, xj, vj, Lj = max(self.points)
            deltai = i - j
            assert deltai > 0, ("should be extrapolating", i, j)
        else:
            j, xj, vj, Lj = min(self.points)
            deltai = i - j
            assert deltai < 0, ("should be extrapolating", i, j)

        newpoint = extrapolate_ahead(deltai, xj, vj)
        return newpoint


class ContourSamplingPath(object):
    """Region-aware form of the sampling path.

    Uses region points to guess a likelihood contour gradient.
    """

    def __init__(self, samplingpath, region):
        """Initialise with `samplingpath` and `region`."""
        self.samplingpath = samplingpath
        self.points = self.samplingpath.points
        self.region = region

    def add(self, i, x, v, L):
        """Add point `xi`, direction `vi` and value `Li` with index `i` to the path."""
        self.samplingpath.add(i, x, v, L)

    def interpolate(self, i):
        """Interpolate point with index `i` on path."""
        return interpolate(
            i, self.samplingpath.points,
            fwd_possible=self.samplingpath.fwd_possible,
            rwd_possible=self.samplingpath.rwd_possible,
            contourpath=self)

    def extrapolate(self, i):
        """Advance beyond the current path, extrapolate from the end point.

        Parameters
        -----------
        i: int
            index on path.

        returns
        --------
        coords: vector
            coordinates of the new point.

        """
        if i >= 0:
            j, xj, vj, Lj = max(self.samplingpath.points)
            deltai = i - j
            assert deltai > 0, ("should be extrapolating", i, j)
        else:
            j, xj, vj, Lj = min(self.samplingpath.points)
            deltai = i - j
            assert deltai < 0, ("should be extrapolating", i, j)

        newpoint = extrapolate_ahead(deltai, xj, vj, contourpath=self)
        return newpoint

    def gradient(self, reflpoint, plot=False):
        """Compute gradient approximation.

        Finds spheres enclosing the `reflpoint`, and chooses their mean
        as the direction to go towards. If no spheres enclose the
        reflpoint, use nearest sphere.

        v is not used, because that would break detailed balance.

        Considerations:
           - in low-d, we want to focus on nearby live point spheres
             The border traced out is fairly accurate, at least in the
             normal away from the inside.

           - in high-d, reflpoint is contained by all live points,
             and none of them point to the center well. Because the
             sampling is poor, the "region center" position
             will be very stochastic.

        Parameters
        -----------
        reflpoint: vector
            point outside the likelihood contour, reflect there
        v: vector
            previous direction vector

        Returns
        ---------
        gradient: vector
            normal of ellipsoid

        """
        if plot:
            plt.plot(reflpoint[0], reflpoint[1], '+ ', color='k', ms=10)

        # check which the reflections the ellipses would make
        region = self.region
        bpts = region.transformLayer.transform(reflpoint.reshape((1,-1)))
        dist = ((bpts - region.unormed)**2).sum(axis=1)
        assert dist.shape == (len(region.unormed),), (dist.shape, len(region.unormed))
        nearby = dist < region.maxradiussq
        assert nearby.shape == (len(region.unormed),), (nearby.shape, len(region.unormed))
        if not nearby.any():
            nearby = dist == dist.min()
        sphere_centers = region.u[nearby,:]

        tsphere_centers = region.unormed[nearby,:]
        nlive, ndim = region.unormed.shape
        assert tsphere_centers.shape[1] == ndim, (tsphere_centers.shape, ndim)

        # choose mean among those points
        tsphere_center = tsphere_centers.mean(axis=0)
        assert tsphere_center.shape == (ndim,), (tsphere_center.shape, ndim)
        tt = get_sphere_tangent(tsphere_center, bpts.flatten())
        assert tt.shape == tsphere_center.shape, (tt.shape, tsphere_center.shape)

        # convert back to u space
        sphere_center = region.transformLayer.untransform(tsphere_center)
        t = region.transformLayer.untransform(tt * 1e-3 + tsphere_center) - sphere_center

        if plot:
            tt_all = get_sphere_tangent(tsphere_centers, bpts)
            t_all = region.transformLayer.untransform(tt_all * 1e-3 + tsphere_centers) - sphere_centers

            """
            # plot in transformed space too:
            origax = plt.gca()
            ax = plt.axes([0.7, 0.7, 0.27, 0.27])
            plt.plot(bpts[:,0], bpts[:,1], '+ ', color='k', ms=10)
            plt.plot(region.unormed[:,0], region.unormed[:,1], 'x ', color='k', ms=4)
            plt.plot(tsphere_centers[:,0], tsphere_centers[:,1], 'o ', mfc='None', mec='b', ms=10, mew=1)
            for si, ti in zip(tsphere_centers, tt_all):
                plt.plot([si[0], ti[0] + si[0]], [si[1], ti[1] + si[1]], '--', lw=2, color='gray', alpha=0.5)
            plt.plot(tsphere_center[0], tsphere_center[1], '^ ', mfc='None', mec='g', ms=10, mew=3)
            plt.plot([tsphere_center[0], tt[0] + tsphere_center[0]],
                [tsphere_center[1], tt[1] + tsphere_center[1]], lw=1, color='gray')
            plt.sca(origax)
            """

            plt.plot(sphere_centers[:,0], sphere_centers[:,1], 'o ', mfc='None', mec='b', ms=10, mew=1)
            if not (dist < region.maxradiussq).any():
                plt.plot(sphere_centers[:,0], sphere_centers[:,1], 's ', mfc='None', mec='b', ms=10, mew=1)
            for si, ti in zip(sphere_centers, t_all):
                plt.plot([si[0], ti[0] * 1000 + si[0]], [si[1], ti[1] * 1000 + si[1]], '--', lw=2, color='gray', alpha=0.5)
            plt.plot(sphere_center[0], sphere_center[1], '^ ', mfc='None', mec='g', ms=10, mew=3)
            plt.plot([sphere_center[0], t[0] * 1000 + sphere_center[0]], [sphere_center[1], t[1] * 1000 + sphere_center[1]], color='gray')

        # compute new vector
        normal = t / norm(t)
        isunitlength(normal)
        assert normal.shape == t.shape, (normal.shape, t.shape)

        return normal



# cython: language_level=3,annotate=True,profile=True,fast_fail=True,warning_errors=True
"""
Region construction methods
---------------------------

Construct and sample from regions of neighbourhoods around the live points.
Includes

* an efficient implementation of MLFriends, with transformation layers and clustering.
  * RadFriends: Buchner (2014) https://arxiv.org/abs/1407.5459
  * MLFriends: Buchner (2019) https://arxiv.org/abs/1707.04476
* a single-ellipsoid region (Mukherjee et al., 2006, https://arxiv.org/abs/astro-ph/0508461)
* a very fast single-ellipsoid, axis-aligned region, for use with step-samplers in high dimensions

"""

import numpy as np
cimport numpy as np
from numpy import pi
cimport cython
from cython.cimports.libc.math import sqrt


@cython.boundscheck(False)
@cython.wraparound(False)
cdef count_nearby(
    np.ndarray[np.float_t, ndim=2] apts,
    np.ndarray[np.float_t, ndim=2] bpts,
    np.float_t radiussq,
    np.ndarray[np.int_t, ndim=1] nnearby
):
    """Count the number of points in ``apts`` within square radius ``radiussq`` for each point ``b`` in `bpts``.

    The number is written to ``nnearby`` (of same length as ``bpts``).

    Parameters
    ----------
    apts: array
        points
    bpts: array
        points
    radiussq: float
        square of the MLFriends radius
    nnearby: array of ints
        The result will be written here.
    """
    cdef size_t na = apts.shape[0]
    cdef size_t nb = bpts.shape[0]
    cdef size_t ndim = apts.shape[1]

    cdef unsigned long i, j
    cdef np.float_t d

    # go through the unselected points and find the worst case
    for j in range(nb):
        # find the nearest selected point
        nnearby[j] = 0
        for i in range(na):
            d = 0
            for k in range(ndim):
                d += (apts[i,k] - bpts[j,k])**2
            if d <= radiussq:
                nnearby[j] += 1


@cython.boundscheck(False)
@cython.wraparound(False)
def find_nearby(
    np.ndarray[np.float_t, ndim=2] apts,
    np.ndarray[np.float_t, ndim=2] bpts,
    np.float_t radiussq,
    np.ndarray[np.int_t, ndim=1] nnearby
):
    """Gets the index of a point in `a` within square radius `radiussq`, for each point `b` in `bpts`.

    The number is written to `nnearby` (of same length as `bpts`).
    If none is found, -1 is written.

    Parameters
    ----------
    apts: array
        points
    bpts: array
        points
    radiussq: float
        square of the MLFriends radius
    nnearby: array of ints
        The result will be written here.

    """
    cdef size_t na = apts.shape[0]
    cdef size_t nb = bpts.shape[0]
    cdef size_t ndim = apts.shape[1]

    cdef unsigned long i, j
    cdef np.float_t d

    # go through the unselected points and find the worst case
    for j in range(nb):
        # find the nearest selected point
        nnearby[j] = -1
        for i in range(na):
            d = 0.0
            for k in range(ndim):
                d += (apts[i,k] - bpts[j,k])**2
            if d <= radiussq:
                nnearby[j] = i
                break


@cython.boundscheck(False)
@cython.wraparound(False)
cdef float compute_maxradiussq(np.ndarray[np.float_t, ndim=2] apts, np.ndarray[np.float_t, ndim=2] bpts):
    """Measure the euclidean distance to the nearest point in `apts`, for each point `b` in `bpts`.

    Parameters
    ----------
    apts: array
        points
    bpts: array
        points

    Returns
    ---------
    array containing the square euclidean distance for each point in bpts.
    """
    cdef size_t na = apts.shape[0]
    cdef size_t nb = bpts.shape[0]
    cdef size_t ndim = apts.shape[1]

    cdef unsigned long i, j
    cdef np.float_t d
    cdef np.float_t mind = 1e300
    cdef np.float_t maxd = 0

    # go through the unselected points and find the worst case
    for j in range(nb):

        # find the nearest selected point
        mind = 1e300
        for i in range(na):
            d = 0
            for k in range(ndim):
                d += (apts[i,k] - bpts[j,k])**2
            mind = min(mind, d)

        maxd = max(maxd, mind)

    return maxd


@cython.boundscheck(False)
@cython.wraparound(False)
def compute_mean_pair_distance(
    np.ndarray[np.float_t, ndim=2] pts,
    np.ndarray[np.int_t, ndim=1] clusterids
):
    """Compute the average distance between pairs of points.
    Pairs from different clusters are excluded in the computation.

    Parameters
    ----------
    pts: array
        points
    clusterids: array of ints or None
        for each point, index of the associated cluster.

    Returns
    ---------
    mean distance between point pairs.
    """
    cdef size_t na = pts.shape[0]
    cdef size_t ndim = pts.shape[1]

    cdef unsigned long i, j
    cdef np.float_t total_dist = 0.0
    cdef np.float_t pair_dist
    cdef int Npairs = 0

    # go through the unselected points and find the worst case
    for j in range(na):
        if clusterids[j] == 0:
            continue
        # find the nearest selected point
        for i in range(j):
            # only consider points in the same cluster
            if clusterids[j] == clusterids[i]:
                pair_dist = 0.0
                for k in range(ndim):
                    pair_dist += (pts[i,k] - pts[j,k])**2
                total_dist += sqrt(pair_dist)
                Npairs += 1

    assert np.isfinite(total_dist), total_dist
    return total_dist / Npairs


@cython.boundscheck(False)
@cython.wraparound(False)
cdef _update_clusters(
    np.ndarray[np.float_t, ndim=2] upoints,
    np.ndarray[np.float_t, ndim=2] tpoints,
    np.float_t maxradiussq,
    np.ndarray[np.int_t, ndim=1] clusterids,
):
    """same signature as ``update_clusters()``, see there."""
    assert upoints.shape[0] == tpoints.shape[0], ('different number of points', upoints.shape[0], tpoints.shape[0])
    assert upoints.shape[1] == tpoints.shape[1], ('different dimensionality of points', upoints.shape[1], tpoints.shape[1])
    clusteridxs = np.zeros(len(tpoints), dtype=int)
    currentclusterid = 1
    i = 0
    # avoid issues when old clusterids are from a longer array
    clusterids = clusterids[:len(tpoints)]
    existing = clusterids == currentclusterid
    if existing.any():
        i = np.where(existing)[0][0]

    clusteridxs[i] = currentclusterid
    while True:
        # compare known members to unassociated
        nonmembermask = clusteridxs == 0
        if not nonmembermask.any():
            # everyone has been assigned -> done!
            break

        nonmembers = tpoints[nonmembermask,:]
        idnearby = np.empty(len(nonmembers), dtype=int)
        members = tpoints[clusteridxs == currentclusterid,:]
        find_nearby(members, nonmembers, maxradiussq, idnearby)
        # print('merging %d into cluster %d of size %d' % (np.count_nonzero(nnearby), currentclusterid, len(members)))

        if (idnearby >= 0).any():
            # place into cluster
            newmembers = nonmembermask
            newmembers[nonmembermask] = idnearby >= 0
            # print('adding', newmembers.sum())
            clusteridxs[newmembers] = currentclusterid
        else:
            # start a new cluster
            currentclusterid = currentclusterid + 1
            i = np.where(nonmembermask)[0][0]
            if clusterids is not None:
                existing = clusterids == currentclusterid
                if existing.any():
                    i = np.where(existing)[0][0]

            clusteridxs[i] = currentclusterid

    assert (clusteridxs > 0).all()
    nclusters = len(np.unique(clusteridxs))
    # assert np.all(np.unique(clusteridxs) == np.arange(nclusters)+1), (np.unique(clusteridxs), nclusters, np.arange(nclusters)+1)
    if nclusters == 1:
        overlapped_upoints = upoints
    else:
        overlapped_upoints = np.empty_like(upoints)
        for idx in np.unique(clusteridxs):
            group_upoints = upoints[clusteridxs == idx,:]
            if len(group_upoints) > 1:
                # center on group mean
                group_mean = group_upoints.mean(axis=0).reshape((1,-1))
            else:
                # if a single point, the differences would be zero,
                # not giving weight to this being an outlier.
                # so use the mean of the entire point population instead
                group_mean = upoints.mean(axis=0).reshape((1,-1))
            overlapped_upoints[clusteridxs == idx,:] = group_upoints - group_mean

    return nclusters, clusteridxs, overlapped_upoints


@cython.boundscheck(False)
@cython.wraparound(False)
def update_clusters(
    np.ndarray[np.float_t, ndim=2] upoints,
    np.ndarray[np.float_t, ndim=2] tpoints,
    np.float_t maxradiussq,
    clusterids=None,
):
    """Clusters `upoints`, so that clusters are distinct if no
    member pair is within a radius of sqrt(`maxradiussq`).

    Parameters
    ----------
    upoints: array
        points (in u-space)
    tpoints: array
        points (in t-space)
    maxradiussq: float
        square of the MLFriends radius
    clusterids: array of ints or None
        for each point, index of the associated cluster.

    Returns
    ---------
    nclusters: int
        the number of clusters found, which is also clusterids.max()
    new_clusterids: array of int
        the new clusterids for each point
    overlapped_points:
        upoints with their cluster centers subtracted.

    The existing cluster ids are re-used when assigning new clusters,
    if possible.
    Clustering is performed on a transformed coordinate space (`tpoints`).
    Returned values are based on upoints.
    """
    if clusterids is None:
        clusterids = np.zeros(len(tpoints), dtype=int)
    return _update_clusters(upoints, tpoints, maxradiussq, clusterids)


@cython.boundscheck(False)
@cython.wraparound(False)
def make_eigvals_positive(
    np.ndarray[np.float_t, ndim=2] a,
    np.float_t targetprod
):
    """For the symmetric square matrix ``a``, increase any zero eigenvalues
    to fulfill a target product of eigenvalues.

    Parameters
    ----------
    a: array
        covariance matrix
    targetprod: array
        target product of eigenvalues

    Returns
    ---------
    covariance matrix
    """

    assert np.isfinite(a).all(), a
    try:
        w, v = np.linalg.eigh(a)  # Use eigh because we assume a is symmetric.
    except np.linalg.LinAlgError as e:
        print(a, targetprod)
        raise e
    mask = w < max(1.e-10, 1e-300**(1. / len(a)))
    if np.any(mask):
        nzprod = np.product(w[~mask])  # product of nonzero eigenvalues
        nzeros = mask.sum()  # number of zero eigenvalues
        w[mask] = (targetprod / nzprod) ** (1. / nzeros)  # adjust zero eigvals
        a = np.dot(np.dot(v, np.diag(w)), np.linalg.inv(v))  # re-form cov

    return a


@cython.boundscheck(False)
@cython.wraparound(False)
def bounding_ellipsoid(
    np.ndarray[np.float_t, ndim=2] x,
    np.float_t minvol=0.
):
    """Calculate bounding ellipsoid containing a set of points x.

    Parameters
    ----------
    x : (npoints, ndim) ndarray
        Coordinates of uniformly sampled points.
    pointvol : float, optional
        Used to set a minimum bound on the ellipsoid volume when
        minvol is True.

    Returns
    -------
    mean and covariance of points

    """
    # Function taken from nestle, MIT licensed, (C) kbarbary

    ndim = x.shape[1]

    # Calculate covariance of points
    ctr = np.mean(x, axis=0)
    delta = x - ctr
    cov = np.cov(delta, rowvar=0)
    assert np.isfinite(cov).all(), (cov, x)
    if ndim == 1:
        cov = np.atleast_2d(cov)

    # For a ball of uniformly distributed points, the covariance will be
    # smaller than r^2 by a factor of 1/(n+2) [see, e.g.,
    # http://mathoverflow.net/questions/35276/
    # covariance-of-points-distributed-in-a-n-ball]. In nested sampling,
    # we are supposing the points are uniformly distributed within
    # an ellipse, so the same factor holds. Expand `cov`
    # to compensate for that when defining the ellipse matrix:
    cov *= (ndim + 2)

    # Ensure that ``cov`` is nonsingular.
    # It can be singular when the ellipsoid has zero volume, which happens
    # when npoints <= ndim or when enough points are linear combinations
    # of other points. (e.g., npoints = ndim+1 but one point is a linear
    # combination of others). When this happens, we expand the ellipse
    # in the zero dimensions to fulfill the volume expected from
    # ``pointvol``.
    if minvol > 0:
        cov = make_eigvals_positive(cov, minvol)

    return ctr, cov


class ScalingLayer(object):
    """Simple transformation layer that only shifts and scales each axis."""

    def __init__(self, mean=0, std=1, nclusters=1, wrapped_dims=[], clusterids=None):
        """Initialise layer."""
        self.mean = mean
        self.std = std
        self.nclusters = nclusters
        self.wrapped_dims = wrapped_dims
        self.has_wraps = len(wrapped_dims) > 0
        self.clusterids = clusterids

    def optimize_wrap(self, points):
        """Optimization for wrapped/circular parameters.
        Does nothing if there are no wrapped/circular parameters.

        Parameters
        ----------
        points: array
            points to use for optimization (in u-space)

        Find largest gap in live points and wrap parameter space there.

        For example, if a wrapped axis has::

            |*****           ****|

        it would identify the middle, subtract it, so that the new space
        is::

            |      ********      |
        """
        if not self.has_wraps:
            return

        N, ndims = points.shape
        self.wrap_cuts = []
        for i in self.wrapped_dims:
            # find largest gap
            vals = np.pad(points[:,i], 1, mode='constant', constant_values=(0,1))
            vals.sort()
            assert vals[0] == 0
            assert vals[-1] == 1
            deltas = vals[1:] - vals[:-1]
            j = deltas.argmax()

            # wrap between i, i+1
            cut = (vals[j] + vals[j+1]) / 2.
            self.wrap_cuts.append(cut)

    def wrap(self, points):
        """Wrap points for circular parameters."""
        if not self.has_wraps:
            return points
        wpoints = points.copy().reshape((-1, points.shape[-1]))
        for i, cut in zip(self.wrapped_dims, self.wrap_cuts):
            wpoints[:,i] = np.fmod(wpoints[:,i] + (1 - cut), 1)
        return wpoints

    def unwrap(self, wpoints):
        """Undo wrapping for circular parameters."""
        if not self.has_wraps:
            return wpoints
        points = wpoints.copy().reshape((-1, wpoints.shape[-1]))
        for i, cut in zip(self.wrapped_dims, self.wrap_cuts):
            points[:,i] = np.fmod(points[:,i] + cut, 1)
        return points

    def optimize(self, points, centered_points, clusterids=None, minvol=0.):
        """Optimize layer.

        Estimates mean and std of points.

        Parameters
        ----------
        points: array
            points to use for optimize (in u-space)
        centered_points: array
            points with their cluster center subtracted
        clusterids: array of ints
            for each point, which cluster they belong to
        minvol:
            ignored
        """
        self.optimize_wrap(points)
        wrapped_points = self.wrap(points)
        self.mean = wrapped_points.mean(axis=0).reshape((1,-1))
        self.std = centered_points.std(axis=0).reshape((1,-1))
        self.axes = np.diag(self.std[0])
        self.logvolscale = np.sum(np.log(self.std))
        self.set_clusterids(clusterids=clusterids, npoints=len(points))

    def set_clusterids(self, clusterids=None, npoints=None):
        """Updates the cluster id assigned to each point."""
        if clusterids is None and self.clusterids is None and npoints is not None:
            # for the beginning, set cluster ids to one for all points
            clusterids = np.ones(npoints, dtype=int)
        if clusterids is not None:
            # if we have a value, update
            self.clusterids = clusterids

    def create_new(self, upoints, maxradiussq, minvol=0.):
        """Learn next layer from this optimized layer's clustering.

        Parameters
        ----------
        upoints: array
            points to use for optimize (in u-space)
        maxradiussq: float
            square of the MLFriends radius
        minvol: float
            Minimum volume to regularize sample covariance

        Returns
        ---------
        A new, optimized ScalingLayer.
        """
        # perform clustering in transformed space
        uwpoints = self.wrap(upoints)
        tpoints = self.transform(upoints)
        nclusters, clusteridxs, overlapped_uwpoints = update_clusters(uwpoints, tpoints, maxradiussq, self.clusterids)
        # clusteridxs = track_clusters(clusteridxs, self.clusterids)
        s = ScalingLayer(nclusters=nclusters, wrapped_dims=self.wrapped_dims, clusterids=clusteridxs)
        s.optimize(upoints, overlapped_uwpoints)
        return s

    def transform(self, u):
        """Transform points from cube space to a whitened space."""
        if self.has_wraps:
            w = self.wrap(u)
        else:
            w = u
        return ((w - self.mean) / self.std).reshape(u.shape)

    def untransform(self, ww):
        """Transform points from whitened space back to cube space."""
        w = (ww * self.std) + self.mean
        if self.has_wraps:
            u = self.unwrap(w).reshape(ww.shape)
        else:
            u = w.reshape(ww.shape)
        return u


class AffineLayer(ScalingLayer):
    """Affine whitening transformation.

    Learns the covariance of points.

    For learning the next layer's covariance, the clustering
    is considered: the sample covariance is computed after subtracting
    the cluster mean. This co-centers all clusters and gets
    the average cluster shape, avoiding learning a covariance dominated
    by the distance between clusters.
    """

    def __init__(self, ctr=0, T=1, invT=1, nclusters=1, wrapped_dims=[], clusterids=None):
        """Initialise layer.

        The parameters are optional and can be learned from points with :meth:`optimize`

        Parameters
        ----------
        ctr: vector
            Center of points
        T: matrix
            Transformation matrix. This matrix whitens the points 
            to a unit Gaussian.
        invT: matrix
            Inverse transformation matrix. For transforming a unit
            Gaussian into something with the sample cov.
        nclusters: int
            number of clusters
        wrapped_dims: array of bools
            indicates which parameter axes are circular.
        clusterids: array of int
            cluster id for each point

        """
        self.ctr = ctr
        self.T = T
        self.invT = invT
        self.nclusters = nclusters
        self.wrapped_dims = wrapped_dims
        self.has_wraps = len(wrapped_dims) > 0
        self.clusterids = clusterids

    def optimize(self, points, centered_points, clusterids=None, minvol=0.):
        """Optimize layer.

        Estimates covariance of ``centered_points``. ``minvol`` sets the
        smallest allowed size of the covariance to avoid numerical
        collapse.

        Parameters
        ----------
        points: array
            points to use for optimize (in u-space)
        centered_points: array
            points with their cluster center subtracted
        clusterids: array of ints
            for each point, which cluster they belong to
        minvol: float
            Minimum volume to regularize sample covariance
        """
        self.optimize_wrap(points)
        wrapped_points = self.wrap(points)
        # point center
        self.ctr = np.mean(wrapped_points, axis=0)
        # compute sample covariance
        cov = np.cov(centered_points, rowvar=0)
        cov *= (len(self.ctr) + 2)
        self.cov = cov
        # Eigen decomposition of the covariance, with numerical stability
        eigval, eigvec = np.linalg.eigh(cov)
        eigvalmin = eigval.max() * 1e-40
        eigval[eigval < eigvalmin] = eigvalmin
        # Try explicit inversion; if this fails, the error is escalated.
        a = np.linalg.inv(cov)
        # log-volume of the space
        self.logvolscale = np.linalg.slogdet(a)[1] * -0.5

        # Transformation matrix with the correct scale
        # this matrix whitens the points to a unit Gaussian.
        self.T = eigvec * eigval**-0.5
        # Inverse transformation matrix, for transforming a unit 
        # Gaussian into something with the sample cov.
        self.invT = np.linalg.inv(self.T)
        # These also are the principle axes of the space
        self.axes = self.invT
        # print('transform used:', self.T, self.invT, 'cov:', cov, 'eigen:', eigval, eigvec)
        self.set_clusterids(clusterids=clusterids, npoints=len(points))

    def create_new(self, upoints, maxradiussq, minvol=0.):
        """Learn next layer from this optimized layer's clustering.

        Parameters
        ----------
        upoints: array
            points to use for optimize (in u-space)
        maxradiussq: float
            square of the MLFriends radius
        minvol: float
            Minimum volume to regularize sample covariance

        Returns
        ---------
        A new, optimized AffineLayer.
        """
        # perform clustering in transformed space
        uwpoints = self.wrap(upoints)
        tpoints = self.transform(upoints)
        nclusters, clusteridxs, overlapped_uwpoints = update_clusters(uwpoints, tpoints, maxradiussq, self.clusterids)
        # clusteridxs = track_clusters(clusteridxs, self.clusterids)
        s = AffineLayer(nclusters=nclusters, wrapped_dims=self.wrapped_dims, clusterids=clusteridxs)
        s.optimize(upoints, overlapped_uwpoints, minvol=minvol)
        return s

    def transform(self, u):
        """Transform points from cube space to a whitened space."""
        if self.has_wraps:
            w = self.wrap(u)
        else:
            w = u
        return np.dot(w - self.ctr, self.T)

    def untransform(self, ww):
        """Transform points from whitened space back to cube space."""
        w = np.dot(ww, self.invT) + self.ctr
        if self.has_wraps:
            u = self.unwrap(w).reshape(ww.shape)
        else:
            u = w.reshape(ww.shape)
        return u

class MaxPrincipleGapAffineLayer(AffineLayer):
    """Affine whitening transformation.

    For learning the next layer's covariance, the clustering
    and principal axis is considered: 
    the sample covariance is computed after subtracting
    the cluster mean. All points are projected onto the line
    defined by the principle axis vector starting from the origin.
    Then, on the sorted line positions, the largest gap is identified.
    All points before the gap are mean-subtracted, and all points
    after the gap are mean-subtracted. Then, the final
    sample covariance is computed. This should give a more "local"
    covariance, even in the case where clusters could not yet be 
    clearly identified.
    """

    def create_new(self, upoints, maxradiussq, minvol=0.):
        """Learn next layer from this optimized layer's clustering.

        Parameters
        ----------
        upoints: array
            points to use for optimize (in u-space)
        maxradiussq: float
            square of the MLFriends radius
        minvol: float
            Minimum volume to regularize sample covariance

        Returns
        ---------
        A new, optimized MaxPrincipleGapAffineLayer.
        """
        # perform clustering in transformed space
        uwpoints = self.wrap(upoints)
        tpoints = self.transform(upoints)
        nclusters, clusteridxs, overlapped_uwpoints = update_clusters(uwpoints, tpoints, maxradiussq, self.clusterids)

        cov = np.cov(overlapped_uwpoints, rowvar=0)
        cov *= (len(self.ctr) + 2)
        eigval, eigvec = np.linalg.eigh(cov)
        # identify principle axis
        principal_vector = eigvec[:, -1]
        # project all overlapped_uwpoints onto principle axis,
        # obtaining position on line
        t = np.dot(overlapped_uwpoints - overlapped_uwpoints.mean(axis=0).reshape((1,-1)), principal_vector)
        # sort positions, identify largest gap
        tsorted = np.sort(t)
        tgapindex = np.argmax(np.diff(tsorted))
        # compute center of largest gap
        tsep = (tsorted[tgapindex] + tsorted[tgapindex + 1]) / 2
        # assign point to left and right cluster
        left_cluster = t < tsep
        # subtract the respective cluster mean from overlapped_uwpoints
        left_mean = overlapped_uwpoints[left_cluster, :].mean(axis=0)
        right_mean = overlapped_uwpoints[~left_cluster, :].mean(axis=0)
        halved_overlapped_uwpoints = overlapped_uwpoints.copy()
        halved_overlapped_uwpoints[left_cluster, :] -= left_mean
        halved_overlapped_uwpoints[~left_cluster, :] -= right_mean

        # re-optimize with the new subtracted points
        s = MaxPrincipleGapAffineLayer(nclusters=nclusters, wrapped_dims=self.wrapped_dims, clusterids=clusteridxs)
        s.optimize(upoints, halved_overlapped_uwpoints, minvol=minvol)
        return s



def vol_prefactor(np.int_t n):
    """Volume constant for an ``n``-dimensional sphere.

    for ``n`` even:  $$    (2pi)^(n    /2) / (2 * 4 * ... * n)$$
    for ``n`` odd :  $$2 * (2pi)^((n-1)/2) / (1 * 3 * ... * n)$$

    Parameters
    ----------
    n: int
        dimensionality

    Returns
    ---------
    volume (float)
    """
    if n % 2 == 0:
        f = 1.
        i = 2
    else:
        f = 2.
        i = 3

    while i <= n:
        f *= 2. / i * pi
        i += 2

    return f


def _inside_ellipsoid(
    np.ndarray[np.float_t, ndim=2] points,
    np.ndarray[np.float_t, ndim=1] ellipsoid_center,
    np.ndarray[np.float_t, ndim=2] ellipsoid_invcov,
    np.float_t square_radius
):
    """Check if inside ellipsoid

    Parameters
    ----------
    points: array of vectors
        Points to check
    ellipsoid_center: vector
        center of ellipsoid
    ellipsoid_invcov: matrix
        inverse covariance matrix
    square_radius: float
        square radius

    Returns
    ---------
    is_inside: array of bools
        True if inside wrapping for each point in ``points``.
    """
    # compute distance vector to center
    d = points - ellipsoid_center
    # distance in normalised coordates: vector . matrix . vector
    # where the matrix is the ellipsoid inverse covariance
    r = np.einsum('ij,jk,ik->i', d, ellipsoid_invcov, d)
    # (r <= 1) means inside
    return r <= square_radius


class MLFriends(object):
    """MLFriends region.

    Defines a region around nested sampling live points for

    1. checking whether a proposed point likely also fulfills the
       likelihood constraints
    2. proposing new points.

    Learns geometry of region from existing live points.
    """

    def __init__(self, u, transformLayer):
        """Initialise region.

        Parameters
        -----------
        u: array of vectors
            live points
        transformLayer: ScalingLayer or AffineLayer
            whitening layer

        """
        if not np.logical_and(u > 0, u < 1).all():
            raise ValueError("not all u values are between 0 and 1: %s" % u[~np.logical_and(u > 0, u < 1).all()])

        self.u = u
        self.set_transformLayer(transformLayer)

        self.sampling_methods = [
            self.sample_from_transformed_boundingbox,
            self.sample_from_boundingbox,
            self.sample_from_points,
            self.sample_from_wrapping_ellipsoid
        ]
        self.current_sampling_method = self.sample_from_boundingbox
        self.vol_prefactor = vol_prefactor(self.u.shape[1])

    def estimate_volume(self):
        """Estimate the order of magnitude of the volume around a single point
        given the current transformLayer.

        Does not account for:
        * the number of live points
        * their overlap
        * the intersection with the unit cube borders

        Returns
        -------
        volume: float
            Volume
        """
        r = self.maxradiussq**0.5
        N, ndim = self.u.shape
        # how large is a sphere of size r in untransformed coordinates?
        return self.transformLayer.logvolscale + np.log(r) * ndim  #+ np.log(vol_prefactor(ndim))

    def set_transformLayer(self, transformLayer):
        """Update transformation layer. Invalidates attribute `maxradius`.

        Parameters
        ----------
        transformLayer: ScalingLayer or AffineLayer
            t-space transformation layer

        """
        self.transformLayer = transformLayer
        self.unormed = self.transformLayer.transform(self.u)
        assert np.isfinite(self.unormed).all(), (self.unormed, self.u)
        self.bbox_lo = self.unormed.min(axis=0)
        self.bbox_hi = self.unormed.max(axis=0)
        self.maxradiussq = None

    def compute_maxradiussq(self, nbootstraps=50):
        """Run MLFriends bootstrapping

        Parameters
        ----------
        nbootstraps: int
            number of bootstrapping rounds

        Returns
        -------
        square radius that safely encloses all live points.
        """
        N, ndim = self.u.shape
        selected = np.empty(N, dtype=bool)
        maxd = 0

        for i in range(nbootstraps):
            idx = np.random.randint(N, size=N)
            selected[:] = False
            selected[idx] = True
            a = self.unormed[selected,:]
            b = self.unormed[~selected,:]

            # compute distances from a to b
            maxd = max(maxd, compute_maxradiussq(a, b))

        assert maxd > 0, (maxd, self.u)
        return maxd

    def compute_enlargement(self, nbootstraps=50, minvol=0., rng=np.random):
        """Return MLFriends radius and ellipsoid enlargement using bootstrapping.

        The wrapping ellipsoid covariance is determined in each bootstrap round.

        Parameters
        ----------
        nbootstraps: int
            number of bootstrapping rounds
        minvol: float
            minimum volume to enforce to wrapping ellipsoid
        rng:
            random number generator

        Returns
        -------
        max_distance: float
            square radius of MLFriends algorithm
        max_radius: float
            square radius of enclosing ellipsoid.
        """
        N, ndim = self.u.shape
        assert np.isfinite(self.unormed).all(), self.unormed
        selected = np.empty(N, dtype=bool)
        maxd = 0.0
        maxf = 0.0

        for i in range(nbootstraps):
            idx = rng.randint(N, size=N)
            selected[:] = False
            selected[idx] = True
            if selected.all() or not selected.any():
                continue

            # compute distances from a to b
            maxd = max(maxd, compute_maxradiussq(
                self.unormed[selected,:],
                self.unormed[~selected,:]))

            # compute enlargement of bounding ellipsoid
            ctr, cov = bounding_ellipsoid(self.u[selected,:], minvol=minvol)
            a = np.linalg.inv(cov)  # inverse covariance
            # compute expansion factor
            delta = self.u[~selected,:] - ctr
            #f = np.einsum('...i, ...i', np.tensordot(delta, a, axes=1), delta).max()
            f = np.einsum('ij,jk,ik->i', delta, a, delta).max()
            assert np.isfinite(f), (ctr, cov, self.unormed, f, delta, a)
            if not f > 0:
                raise np.linalg.LinAlgError("Distances are not positive")
            maxf = max(maxf, f)

        assert maxd > 0, (maxd, self.u, self.unormed)
        assert maxf > 0, (maxf, self.u, self.unormed)
        return maxd, maxf

    def sample_from_points(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Chooses randomly from points and their ellipsoids.

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # generate points near random existing points
        idx = np.random.randint(N, size=nsamples)
        v = np.random.normal(size=(nsamples, ndim))
        v *= (np.random.uniform(size=nsamples)**(1. / ndim) / np.linalg.norm(v, axis=1)).reshape((-1, 1))
        v = self.unormed[idx,:] + v * self.maxradiussq**0.5

        # count how many are around
        nnearby = np.empty(nsamples, dtype=int)
        count_nearby(self.unormed, v, self.maxradiussq, nnearby)
        vmask = np.random.uniform(high=nnearby) < 1
        w = self.transformLayer.untransform(v[vmask,:])
        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        wmask[wmask] = self.inside_ellipsoid(w[wmask])

        return w[wmask,:]

    def sample_from_boundingbox(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from bounding box around region.

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # draw from unit cube in prior space
        u = np.random.uniform(size=(nsamples, ndim))
        wmask = self.inside_ellipsoid(u)
        # check if inside region in transformed space
        v = self.transformLayer.transform(u[wmask,:])
        idnearby = np.empty(len(v), dtype=int)
        find_nearby(self.unormed, v, self.maxradiussq, idnearby)
        vmask = idnearby >= 0
        return u[wmask,:][vmask,:]

    def sample_from_transformed_boundingbox(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from bounding box around region (in whitened space).

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # draw from rectangle in transformed space
        v = np.random.uniform(self.bbox_lo - self.maxradiussq, self.bbox_hi + self.maxradiussq, size=(nsamples, ndim))
        idnearby = np.empty(nsamples, dtype=int)
        find_nearby(self.unormed, v, self.maxradiussq, idnearby)
        vmask = idnearby >= 0

        # check if inside unit cube
        w = self.transformLayer.untransform(v[vmask,:])
        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        wmask[wmask] = self.inside_ellipsoid(w[wmask])

        return w[wmask,:]

    def sample_from_wrapping_ellipsoid(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from wrapping ellipsoid and filters with region.

        Parameters as described in ``sample()``.
        """
        N, ndim = self.u.shape
        # draw from rectangle in transformed space

        z = np.random.normal(size=(nsamples, ndim))
        assert ((z**2).sum(axis=1) > 0).all(), (z**2).sum(axis=1)
        z /= ((z**2).sum(axis=1)**0.5).reshape((nsamples, 1))
        assert self.enlarge > 0, self.enlarge
        u = z * self.enlarge**0.5 * np.random.uniform(size=(nsamples, 1))**(1./ndim)

        w = self.ellipsoid_center + np.dot(u, self.ellipsoid_axes_T)
        #assert self.inside_ellipsoid(w).all()

        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        v = self.transformLayer.transform(w[wmask,:])
        idnearby = np.empty(len(v), dtype=int)
        find_nearby(self.unormed, v, self.maxradiussq, idnearby)
        vmask = idnearby >= 0

        return w[wmask,:][vmask,:]

    def sample(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Switches automatically between the ``sampling_methods`` (attribute).

        Parameters
        ----------
        nsamples: int
            number of samples to draw

        Returns
        -------
        samples: array of shape (nsamples, dimension)
            samples drawn
        idx: array of integers (nsamples)
            index of a point nearby (MLFriends.u)
        """
        samples = self.current_sampling_method(nsamples=nsamples)
        if len(samples) == 0:
            # no result, choose another method
            self.current_sampling_method = self.sampling_methods[np.random.randint(len(self.sampling_methods))]
            # print("switching to %s" % self.current_sampling_method)
        return samples

    def inside(self, pts):
        """Check if inside region.

        Parameters
        ----------
        pts: array of vectors
            Points to check

        Returns
        ---------
        is_inside: array of bools
            True if inside MLFriends region and wrapping ellipsoid,
            for each point in ``pts``.

        """
        # require points to be inside bounding ellipsoid
        mask = self.inside_ellipsoid(pts)

        if mask.any():
            # additionally require points to be near neighbours
            bpts = self.transformLayer.transform(pts[mask,:])
            idnearby = np.empty(len(bpts), dtype=int)
            find_nearby(self.unormed, bpts, self.maxradiussq, idnearby)
            mask[mask] = idnearby >= 0

        return mask

    def create_ellipsoid(self, minvol=0.0):
        """Create wrapping ellipsoid and store its center and covariance.

        Parameters
        ----------
        minvol: float
            If positive, make sure ellipsoid has at least this volume.
        """
        assert self.enlarge is not None
        # compute enlargement of bounding ellipsoid
        ctr, cov = bounding_ellipsoid(self.u, minvol=minvol)
        a = np.linalg.inv(cov)

        self.ellipsoid_center = ctr
        self.ellipsoid_invcov = a
        self.ellipsoid_cov = cov

        l, v = np.linalg.eigh(a)
        self.ellipsoid_axlens = 1. / np.sqrt(l)
        self.ellipsoid_axes = np.dot(v, np.diag(self.ellipsoid_axlens))
        self.ellipsoid_axes_T = self.ellipsoid_axes.transpose()

        l2, v2 = np.linalg.eigh(cov)
        self.ellipsoid_inv_axlens = 1. / np.sqrt(l2)
        self.ellipsoid_inv_axes = np.dot(v2, np.diag(self.ellipsoid_inv_axlens))


    def inside_ellipsoid(self, u):
        """Check if inside wrapping ellipsoid.

        Parameters
        ----------
        u: array of vectors
            Points to check

        Returns
        ---------
        is_inside: array of bools
            True if inside wrapping ellipsoid, for each point in `pts`.

        """
        return _inside_ellipsoid(u, self.ellipsoid_center, self.ellipsoid_invcov, self.enlarge)

    def compute_mean_pair_distance(self):
        return compute_mean_pair_distance(self.unormed, self.transformLayer.clusterids)


class RobustEllipsoidRegion(MLFriends):
    """Ellipsoidal region.

    Defines a region around nested sampling live points for

    1. checking whether a proposed point likely also fulfills the
       likelihood constraints
    2. proposing new points.

    Learns geometry of region from existing live points.
    """

    def __init__(self, u, transformLayer):
        """Initialise region.

        Parameters
        -----------
        u: array of vectors
            live points
        transformLayer: ScalingLayer or AffineLayer
            whitening layer

        """
        if not np.logical_and(u > 0, u < 1).all():
            raise ValueError("not all u values are between 0 and 1: %s" % u[~np.logical_and(u > 0, u < 1).all()])

        self.u = u
        self.set_transformLayer(transformLayer)

        self.sampling_methods = [
            #self.sample_from_transformed_boundingbox,
            self.sample_from_boundingbox,
            self.sample_from_wrapping_ellipsoid
        ]
        self.current_sampling_method = self.sample_from_boundingbox
        self.vol_prefactor = vol_prefactor(self.u.shape[1])

    def sample_from_boundingbox(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from bounding box around region.

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # draw from unit cube in prior space
        u = np.random.uniform(size=(nsamples, ndim))
        wmask = self.inside_ellipsoid(u)
        return u[wmask,:]

    def sample_from_transformed_boundingbox(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from bounding box around region (in whitened space).

        Parameters as described in *sample()*.
        """
        N, ndim = self.u.shape
        # draw from rectangle in transformed space
        v = np.random.uniform(self.bbox_lo - self.maxradiussq, self.bbox_hi + self.maxradiussq, size=(nsamples, ndim))

        # check if inside unit cube
        w = self.transformLayer.untransform(v)
        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        wmask[wmask] = self.inside_ellipsoid(w[wmask])

        return w[wmask,:]

    def sample_from_wrapping_ellipsoid(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Draws uniformly from wrapping ellipsoid and filters with region.

        Parameters as described in ``sample()``.
        """
        N, ndim = self.u.shape
        # draw from rectangle in transformed space

        z = np.random.normal(size=(nsamples, ndim))
        assert ((z**2).sum(axis=1) > 0).all(), (z**2).sum(axis=1)
        z /= ((z**2).sum(axis=1)**0.5).reshape((nsamples, 1))
        assert self.enlarge > 0, self.enlarge
        u = z * self.enlarge**0.5 * np.random.uniform(size=(nsamples, 1))**(1./ndim)

        w = self.ellipsoid_center + np.dot(u, self.ellipsoid_axes_T)
        #assert self.inside_ellipsoid(w).all()

        wmask = np.logical_and(w > 0, w < 1).all(axis=1)
        return w[wmask,:]

    def sample(self, nsamples=100):
        """Draw uniformly sampled points from MLFriends region.

        Switches automatically between the ``sampling_methods`` (attribute).

        Parameters
        ----------
        nsamples: int
            number of samples to draw

        Returns
        -------
        samples: array of shape (nsamples, dimension)
            samples drawn
        idx: array of integers (nsamples)
            index of a point nearby (MLFriends.u)
        """
        samples = self.current_sampling_method(nsamples=nsamples)
        if len(samples) == 0:
            # no result, choose another method
            self.current_sampling_method = self.sampling_methods[np.random.randint(len(self.sampling_methods))]
            #print("switching to %s" % self.current_sampling_method)
        return samples

    def inside(self, pts):
        """Check if inside region.

        Parameters
        ----------
        pts: array of vectors
            Points to check

        Returns
        ---------
        is_inside: array of bools
            True if inside MLFriends region and wrapping ellipsoid,
            for each point in ``pts``.

        """
        # require points to be inside bounding ellipsoid
        return self.inside_ellipsoid(pts)

    def compute_enlargement(self, nbootstraps=50, minvol=0., rng=np.random):
        """Return MLFriends radius and ellipsoid enlargement using bootstrapping.

        The wrapping ellipsoid covariance is determined in each bootstrap round.

        Parameters
        ----------
        nbootstraps: int
            number of bootstrapping rounds
        minvol: float
            minimum volume to enforce to wrapping ellipsoid
        rng:
            random number generator

        Returns
        -------
        max_distance: float
            square radius of MLFriends algorithm
        max_radius: float
            square radius of enclosing ellipsoid.
        """
        N, ndim = self.u.shape
        if N < ndim + 1:
            raise FloatingPointError('not enough live points to compute covariance')
        assert np.isfinite(self.unormed).all(), self.unormed
        selected = np.empty(N, dtype=bool)
        maxd = 1e300
        maxf = 0.0

        for i in range(nbootstraps):
            idx = rng.randint(N, size=N)
            selected[:] = False
            selected[idx] = True

            # compute enlargement of bounding ellipsoid
            ctr, cov = bounding_ellipsoid(self.u[selected,:])
            a = np.linalg.inv(cov)  # inverse covariance
            # compute expansion factor
            delta = self.u[~selected,:] - ctr
            #f = np.einsum('...i, ...i', np.tensordot(delta, a, axes=1), delta).max()
            f = np.einsum('ij,jk,ik->i', delta, a, delta).max()
            assert np.isfinite(f), (ctr, cov, self.unormed, f, delta, a)
            if not f > 0:
                raise np.linalg.LinAlgError("Distances are not positive")
            maxf = max(maxf, f)

        assert maxd > 0, (maxd, self.u, self.unormed)
        assert maxf > 0, (maxf, self.u, self.unormed)
        return maxd, maxf

    def estimate_volume(self):
        """Estimate the volume of the ellipsoid.

        Does not account for the intersection with the unit cube borders.

        Returns
        -------
        logvolume: float
            logarithm of the volume.
        """
        ndim = len(self.ellipsoid_cov)
        sign, logvol = np.linalg.slogdet(self.ellipsoid_cov)
        if sign > 0:
            return logvol + ndim * np.log(self.enlarge)
        else:
            return -1e300


class SimpleRegion(RobustEllipsoidRegion):
    """Axis-aligned ellipsoidal region.

    Defines a region around nested sampling live points for

    1. checking whether a proposed point likely also fulfills the
       likelihood constraints
    2. proposing new points.

    Learns geometry of region from existing live points.
    """

    def create_ellipsoid(self, minvol=0.0):
        """Create wrapping ellipsoid and store its center and covariance.

        Parameters
        ----------
        minvol: float
            If positive, make sure ellipsoid has at least this volume.
        """
        assert self.enlarge is not None
        # compute enlargement of bounding ellipsoid
        ctr = np.mean(self.u, axis=0)
        var = np.var(self.u, axis=0)
        a = np.diag(1. / var)
        cov = np.diag(var)

        self.ellipsoid_center = ctr
        self.ellipsoid_invcov = a
        self.ellipsoid_cov = cov

        l, v = np.linalg.eigh(a)
        self.ellipsoid_axlens = 1. / np.sqrt(l)
        self.ellipsoid_axes = np.dot(v, np.diag(self.ellipsoid_axlens))
        self.ellipsoid_axes_T = self.ellipsoid_axes.transpose()

        l2, v2 = np.linalg.eigh(cov)
        self.ellipsoid_inv_axlens = 1. / np.sqrt(l2)
        self.ellipsoid_inv_axes = np.dot(v2, np.diag(self.ellipsoid_inv_axlens))


    def compute_enlargement(self, nbootstraps=50, minvol=0., rng=np.random):
        """Return MLFriends radius and ellipsoid enlargement using bootstrapping.

        The wrapping ellipsoid covariance is determined in each bootstrap round.

        Parameters
        ----------
        nbootstraps: int
            number of bootstrapping rounds
        minvol: float
            minimum volume to enforce to wrapping ellipsoid
        rng:
            random number generator

        Returns
        -------
        max_distance: float
            square radius of MLFriends algorithm
        max_radius: float
            square radius of enclosing ellipsoid.
        """
        N, ndim = self.u.shape
        assert np.isfinite(self.u).all(), self.u
        assert np.isfinite(self.unormed).all(), self.unormed
        selected = np.empty(N, dtype=bool)
        maxd = 1e300
        maxf = 0.0
        if N < ndim + 1:
            raise FloatingPointError('not enough live points to compute variance')

        for i in range(nbootstraps):
            idx = rng.randint(N, size=N)
            selected[:] = False
            selected[idx] = True

            # compute enlargement of bounding ellipsoid
            ctr = np.mean(self.u[selected,:], axis=0)
            var = np.var(self.u[selected,:], axis=0)
            # compute expansion factor
            f = np.sum((self.u[~selected,:] - ctr.reshape((1, -1)))**2 / var, axis=0).max()
            assert np.isfinite(f), (self.u, ctr, var, self.unormed, f)
            if not f > 0:
                raise np.linalg.LinAlgError("Distances are not positive")
            maxf = max(maxf, f)

        assert maxd > 0, (maxd, self.u, self.unormed)
        assert maxf > 0, (maxf, self.u, self.unormed)
        return maxd, maxf


class WrappingEllipsoid(object):
    """Ellipsoid which safely wraps points."""

    def __init__(self, u):
        """Initialise region.

        Parameters
        -----------
        u: array of vectors
            live points
        """
        self.u = u
        # allow some parameters to have exactly the same value
        # this can occur with grid / categorical parameters
        self.variable_dims = np.std(self.u, axis=0) > 0
        if self.variable_dims.all():
            self.variable_dims = Ellipsis

    def compute_enlargement(self, nbootstraps=50, rng=np.random):
        """Return ellipsoid enlargement after `nbootstraps` bootstrapping rounds.

        The wrapping ellipsoid covariance is determined in each bootstrap round.
        """
        N = len(self.u)
        v = self.u[:,self.variable_dims]
        selected = np.empty(N, dtype=bool)
        maxf = 0.0

        for i in range(nbootstraps):
            idx = rng.randint(N, size=N)
            selected[:] = False
            selected[idx] = True
            ua = v[selected,:]
            ub = v[~selected,:]

            # compute enlargement of bounding ellipsoid
            ctr, cov = bounding_ellipsoid(ua)
            a = np.linalg.inv(cov)  # inverse covariance
            # compute expansion factor
            delta = ub - ctr
            f = np.einsum('...i, ...i', np.tensordot(delta, a, axes=1), delta).max()
            if not f > 0:
                raise np.linalg.LinAlgError("Distances are not positive")
            maxf = max(maxf, f)

        assert maxf > 0, (maxf, self.u, self.active_dims)
        return maxf

    def create_ellipsoid(self, minvol=0.0):
        """Create wrapping ellipsoid and store its center and covariance."""
        assert self.enlarge is not None
        # compute enlargement of bounding ellipsoid
        ctr, cov = bounding_ellipsoid(self.u[:,self.variable_dims], minvol=minvol)
        a = np.linalg.inv(cov)

        self.ellipsoid_center = ctr
        self.ellipsoid_invcov = a
        self.ellipsoid_cov = cov

        l, v = np.linalg.eigh(a)
        self.ellipsoid_axlens = 1. / np.sqrt(l)
        self.ellipsoid_axes = np.dot(v, np.diag(self.ellipsoid_axlens))

    def update_center(self, ctr):
        """Update ellipsoid center, considering fixed dimensions.

        Parameters
        ----------
        ctr: vector
            new center

        """
        if self.variable_dims is Ellipsis:
            self.ellipsoid_center = ctr
        else:
            self.ellipsoid_center = ctr[self.variable_dims]

    def inside(self, u):
        """Check if inside wrapping ellipsoid.

        Parameters
        ----------
        u: array of vectors
            Points to check

        Returns
        ---------
        is_inside: array of bools
            True if inside wrapping ellipsoid, for each point in `pts`.

        """
        # check the variable subspace with the ellipsoid
        inside_variable = _inside_ellipsoid(u[:,self.variable_dims], self.ellipsoid_center, self.ellipsoid_invcov, self.enlarge)
        if self.variable_dims is Ellipsis:
            return inside_variable
        else:
            # the remaining dims must be exactly equal
            inside_fixed = np.all(self.u[0, ~self.variable_dims] == u[:,~self.variable_dims], axis=1)
            return np.logical_and(inside_fixed, inside_variable)



"""Drop-in replacement for pymultinest.solve.

Example::

    from ultranest.solvecompat import pymultinest_solve_compat as solve

    # is a drop-in replacement for

    from pymultinest.solve import solve

"""


import numpy as np
import string

from .integrator import ReactiveNestedSampler
from .stepsampler import SliceSampler, generate_mixture_random_direction


def pymultinest_solve_compat(
    LogLikelihood, Prior, n_dims, paramnames=None,
    outputfiles_basename=None, resume=False,
    n_live_points=400, evidence_tolerance=0.5,
    seed=-1, max_iter=0, wrapped_params=None, verbose=True,
    speed="safe",
    **kwargs
):
    """Run nested sampling analysis.

    Disadvantages compared to using ReactiveNestedSampler directly:
    cannot resume easily, cannot plot interactively.
    Limited results.

    It is recommended that you directly use::

        sampler = ReactiveNestedSampler(paramnames, LogLikelihood, transform=Prior)
        sampler.run()

    following the UltraNest documentation and manuals,
    as this gives you more control on resuming and sampler options.
    """
    if paramnames is None:
        paramnames = list(string.ascii_lowercase)[:n_dims]
    if seed >= 0:
        np.random.seed(seed)
    assert len(paramnames) == n_dims
    min_ess = kwargs.pop('min_ess', 0)
    frac_remain = kwargs.pop('frac_remain', 0.01)
    Lepsilon = kwargs.pop('Lepsilon', 0.001)
    outputkwargs = {}
    if not verbose:
        outputkwargs = dict(viz_callback=False, show_status=False)

    sampler = ReactiveNestedSampler(
        paramnames, LogLikelihood, transform=Prior,
        log_dir=outputfiles_basename, resume='resume' if resume else 'overwrite',
        wrapped_params=wrapped_params, draw_multiple=False, vectorized=False,
        **outputkwargs)

    if speed == "safe":
        pass
    elif speed == "auto":
        sampler.run(
            dlogz=evidence_tolerance,
            max_iters=max_iter if max_iter > 0 else None,
            min_num_live_points=n_live_points,
            min_ess=min_ess, frac_remain=frac_remain,
            Lepsilon=Lepsilon, max_ncalls=40000)

        sampler.stepsampler = SliceSampler(
            nsteps=1000,
            generate_direction=generate_mixture_random_direction,
            adaptive_nsteps='move-distance',
            region_filter=kwargs.get('region_filter', True)
        )
    else:
        sampler.stepsampler = SliceSampler(
            generate_direction=generate_mixture_random_direction,
            nsteps=speed,
            adaptive_nsteps=False,
            region_filter=False)

    sampler.run(dlogz=evidence_tolerance,
                max_iters=max_iter if max_iter > 0 else None,
                min_num_live_points=n_live_points,
                min_ess=min_ess, frac_remain=frac_remain,
                Lepsilon=Lepsilon)

    if verbose:
        sampler.print_results()
    results = sampler.results
    sampler.plot()

    return dict(logZ=results['logz'],
                logZerr=results['logzerr'],
                samples=results['samples'],
                weighted_samples=results['weighted_samples'])
